from typing import Optional
import networkx as nx
import numpy
from matchms import Scores
from .networking_functions import get_top_hits


class SimilarityNetwork:
    """Create a spectal network from spectrum similarities.

    For example

    .. testcode::

        import numpy as np
        from matchms import Spectrum, calculate_scores
        from matchms.similarity import ModifiedCosine
        from matchms.networking import SimilarityNetwork

        spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.7, 0.2, 0.1]),
                              metadata={"precursor_mz": 100.0,
                                        "testID": "one"})
        spectrum_2 = Spectrum(mz=np.array([104.9, 140, 190.]),
                              intensities=np.array([0.4, 0.2, 0.1]),
                              metadata={"precursor_mz": 105.0,
                                        "testID": "two"})

        # Use factory to construct a similarity function
        modified_cosine = ModifiedCosine(tolerance=0.2)
        spectrums = [spectrum_1, spectrum_2]
        scores = calculate_scores(spectrums, spectrums, modified_cosine)
        ms_network = SimilarityNetwork(identifier_key="testID")
        ms_network.create_network(scores)

        nodes = list(ms_network.graph.nodes())
        nodes.sort()
        print(nodes)

    Should output

    .. testoutput::

        ['one', 'two']

    """
    def __init__(self, identifier_key: str = "spectrumid",
                 top_n: int = 20,
                 max_links: int = 10,
                 score_cutoff: float = 0.7,
                 link_method: str = 'single',
                 keep_unconnected_nodes: bool = True):
        """
        Parameters
        ----------
        identifier_key
            Metadata key for unique intentifier for each spectrum in scores.
            Will also be used for the naming the network nodes. Default is 'spectrumid'.
        top_n
            Consider edge between spectrumA and spectrumB if score falls into
            top_n for spectrumA or spectrumB (link_method="single"), or into
            top_n for spectrumA and spectrumB (link_method="mutual"). From those
            potential links, only max_links will be kept, so top_n must be >= max_links.
        max_links
            Maximum number of links to add per node. Default = 10.
            Due to incoming links, total number of links per node can be higher.
            The links are populated by looping over the query spectrums.
            Important side note: The max_links restriction is strict which means that
            if scores around max_links are equal still only max_links will be added
            which can results in some random variations (sorting spectra with equal
            scores restuls in a random order of such elements).
        score_cutoff
            Threshold for given similarities. Edges/Links will only be made for
            similarities > score_cutoff. Default = 0.7.
        link_method
            Chose between 'single' and 'mutual'. 'single will add all links based
            on individual nodes. 'mutual' will only add links if that link appears
            in the given top-n list for both nodes.
        keep_unconnected_nodes
            If set to True (default) all spectra will be included as nodes even
            if they have no connections/edges of other spectra. If set to False
            all nodes without connections will be removed.
        """
        # pylint: disable=too-many-arguments
        self.identifier_key = identifier_key
        self.top_n = top_n
        self.max_links = max_links
        self.score_cutoff = score_cutoff
        self.link_method = link_method
        self.keep_unconnected_nodes = keep_unconnected_nodes
        self.graph: Optional[nx.Graph] = None
        """NetworkX graph. Set after calling create_network()"""

    @staticmethod
    def _select_edge_score(similars_scores: dict, scores_type: numpy.dtype):
        """Chose one value if score contains multiple values (e.g. "score" and "matches")"""
        if len(scores_type) > 1 and "score" in scores_type.names:
            return {key: value["score"] for key, value in similars_scores.items()}
        if len(scores_type) > 1:  # Assume that first entry is desired score
            return {key: value[0] for key, value in similars_scores.items()}
        return similars_scores

    def create_network(self, scores: Scores):
        """
        Function to create network from given top-n similarity values. Expects that
        similarities given in scores are from an all-vs-all comparison including all
        possible pairs.

        Parameters
        ----------
        scores
            Matchms Scores object containing all spectrums and pair similarities for
            generating a network.
        """
        assert self.top_n >= self.max_links, "top_n must be >= max_links"
        assert numpy.all(scores.queries == scores.references), \
            "Expected symmetric scores object with queries==references"
        unique_ids = list({s.get(self.identifier_key) for s in scores.queries})

        # Initialize network graph, add nodes
        msnet = nx.Graph()
        msnet.add_nodes_from(unique_ids)

        # Collect location and score of highest scoring candidates for queries and references
        similars_idx, similars_scores = get_top_hits(scores, identifier_key=self.identifier_key,
                                                     top_n=self.top_n,
                                                     search_by="queries",
                                                     ignore_diagonal=True)
        similars_scores = self._select_edge_score(similars_scores, scores.scores.dtype)

        # Add edges based on global threshold (cutoff) for weights
        for i, spec in enumerate(scores.queries):
            query_id = spec.get(self.identifier_key)

            ref_candidates = numpy.array([scores.references[x].get(self.identifier_key)
                                          for x in similars_idx[query_id]])
            idx = numpy.where((similars_scores[query_id] >= self.score_cutoff) &
                              (ref_candidates != query_id))[0][:self.max_links]
            if self.link_method == "single":
                new_edges = [(query_id, str(ref_candidates[x]),
                              float(similars_scores[query_id][x])) for x in idx]
            elif self.link_method == "mutual":
                new_edges = [(query_id, str(ref_candidates[x]),
                              float(similars_scores[query_id][x]))
                             for x in idx if i in similars_idx[ref_candidates[x]][:]]
            else:
                raise ValueError("Link method not kown")

            msnet.add_weighted_edges_from(new_edges)

        if not self.keep_unconnected_nodes:
            msnet.remove_nodes_from(list(nx.isolates(msnet)))
        self.graph = msnet

    def export_to_graphml(self, filename: str):
        """Save the network as .graphml file.

        Parameters
        ----------
        filename
            Specify filename for exporting the graph.

        """
        if not self.graph:
            raise ValueError("No network found. Make sure to first run .create_network() step")
        nx.write_graphml_lxml(self.graph, filename)



import logging
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def require_precursor_below_mz(spectrum_in: SpectrumType, max_mz: float = 1000) -> SpectrumType:

    """Returns None if the precursor_mz of a spectrum is above
       max_mz.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    max_mz:
        Maximum mz value for the precursor mz of a spectrum.
        All precursor mz values greater or equal to this
        will return none. Default is 1000.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    precursor_mz = spectrum.get("precursor_mz", None)
    assert precursor_mz is not None, "Precursor mz absent."
    assert isinstance(precursor_mz, (float, int)), ("Expected 'precursor_mz' to be a scalar number.",
                                                    "Consider applying 'add_precursor_mz' filter first.")
    assert max_mz >= 0, "max_mz must be a positive scalar."
    if precursor_mz >= max_mz:
        logger.info("Spectrum with precursor_mz %s (>%s) was set to None.",
                    str(precursor_mz), str(max_mz))
        return None

    return spectrum



import json
import logging
import numpy as np
import requests
from ..Spectrum import Spectrum


logger = logging.getLogger("matchms")


def load_from_usi(usi: str, server: str = "https://metabolomics-usi.ucsd.edu"):
    """Load spectrum from metabolomics USI.

    USI returns JSON data with keys "peaks", "n_peaks" and "precuror_mz"

    .. code-block:: python

        from matchms.importing import load_from_usi

        spectrum = load_from_usi("mzspec:MASSBANK::accession:SM858102")
        print(f"Found spectrum with precursor m/z of {spectrum.get("precursor_mz"):.2f}.")

    Parameters
    ----------
    usi:
        Provide the usi.

    server: string
        USI server
    """

    # Create the url
    url = server + "/json/?usi1=" + usi
    metadata = {"usi": usi, "server": server}
    response = requests.get(url)

    if response.status_code == 404:
        return None
    # Extract data and create Spectrum object
    try:
        spectral_data = response.json()
        if spectral_data is None or "peaks" not in spectral_data:
            logger.info("Empty spectrum found (no data found). Will not be imported.")
            return None
        peaks = spectral_data["peaks"]
        if len(peaks) == 0:
            logger.info("Empty spectrum found (no peaks in 'peaks_json'). Will not be imported.")
            return None
        mz_list, intensity_list = zip(*peaks)
        mz_array = np.array(mz_list)
        intensity_array = np.array(intensity_list)

        metadata["precursor_mz"] = spectral_data.get("precursor_mz", None)

        s = Spectrum(mz_array, intensity_array, metadata)

        return s

    except json.decoder.JSONDecodeError:
        logger.warning("Failed to unpack json (JSONDecodeError).")
        return None



from abc import abstractmethod
from typing import List
import numpy
from matchms.typing import SpectrumType


class BaseSimilarity:
    """Similarity function base class.
    When building a custom similarity measure, inherit from this class and implement
    the desired methods.

    Attributes
    ----------
    is_commutative
       Whether similarity function is commutative, which means that the order of spectrums
       does not matter (similarity(A, B) == similarity(B, A)). Default is True.
    """
    # Set key characteristics as class attributes
    is_commutative = True
    # Set output data type, e.g. "float" or [("score", "float"), ("matches", "int")]
    score_datatype = numpy.float64

    @abstractmethod
    def pair(self, reference: SpectrumType, query: SpectrumType) -> float:
        """Method to calculate the similarity for one input pair.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.

        Returns
            score as numpy array (using self.score_datatype). For instance returning
            numpy.asarray(score, dtype=self.score_datatype)
        """
        raise NotImplementedError

    def matrix(self, references: List[SpectrumType], queries: List[SpectrumType],
               is_symmetric: bool = False) -> numpy.ndarray:
        """Optional: Provide optimized method to calculate an numpy.array of similarity scores
        for given reference and query spectrums. If no method is added here, the following naive
        implementation (i.e. a double for-loop) is used.

        Parameters
        ----------
        references
            List of reference objects
        queries
            List of query objects
        is_symmetric
            Set to True when *references* and *queries* are identical (as for instance for an all-vs-all
            comparison). By using the fact that score[i,j] = score[j,i] the calculation will be about
            2x faster.
        """
        n_rows = len(references)
        n_cols = len(queries)
        scores = numpy.empty([n_rows, n_cols], dtype=self.score_datatype)
        for i_ref, reference in enumerate(references[:n_rows]):
            if is_symmetric and self.is_commutative:
                for i_query, query in enumerate(queries[i_ref:n_cols], start=i_ref):
                    scores[i_ref][i_query] = self.pair(reference, query)
                    scores[i_query][i_ref] = scores[i_ref][i_query]
            else:
                for i_query, query in enumerate(queries[:n_cols]):
                    scores[i_ref][i_query] = self.pair(reference, query)
        return scores

    def sort(self, scores: numpy.ndarray):
        """Return array of indexes for sorted list of scores.
        This method can be adapted for different styles of scores.

        Parameters
        ----------
        scores
            1D Array of scores.

        Returns
        -------
        idx_sorted
            Indexes of sorted scores.
        """
        return scores.argsort()[::-1]



import logging
from ..typing import SpectrumType
from ..utils import clean_adduct
from ..utils import looks_like_adduct


logger = logging.getLogger("matchms")


def derive_adduct_from_name(spectrum_in: SpectrumType,
                            remove_adduct_from_name: bool = True) -> SpectrumType:
    """Find adduct in compound name and add to metadata (if not present yet).

    Method to interpret the given compound name to find the adduct.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    remove_adduct_from_name:
        Remove found adducts from compound name if set to True. Default is True.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    if spectrum.get("compound_name", None) is not None:
        name = spectrum.get("compound_name")
    else:
        assert spectrum.get("name", None) in [None, ""], ("Found 'name' but not 'compound_name' in metadata",
                                                          "Apply 'add_compound_name' filter first.")
        return spectrum

    # Detect adduct in compound name
    adduct_from_name = None
    name_split = name.split(" ")
    for name_part in name_split[::-1][:2]:
        if looks_like_adduct(name_part):
            adduct_from_name = name_part
            break

    if adduct_from_name and remove_adduct_from_name:
        name_adduct_removed = " ".join([x for x in name_split if x != adduct_from_name])
        name_adduct_removed = name_adduct_removed.strip("; ")
        spectrum.set("compound_name", name_adduct_removed)
        logger.info("Removed adduct %s from compound name.", adduct_from_name)

    # Add found adduct to metadata (if not present yet)
    if adduct_from_name and not looks_like_adduct(spectrum.get("adduct")):
        adduct_cleaned = clean_adduct(adduct_from_name)
        spectrum.set("adduct", adduct_cleaned)
        logger.info("Added adduct %s to metadata.", adduct_cleaned)

    return spectrum



import logging
import os
from typing import IO
from typing import Dict
from typing import List
from typing import Union
from ..Spectrum import Spectrum
from ..Spikes import Spikes


logger = logging.getLogger("matchms")


_extentions_not_allowed = ["mzml", "mzxml", "json", "mgf"]


def save_as_msp(spectra: List[Spectrum], filename: str,
                write_peak_comments: bool = True):
    """Save spectrum(s) as msp file.

    :py:attr:`~matchms.Spectrum.losses` of spectrum will not be saved.

    Example:

    .. code-block:: python

        import numpy
        from matchms import Spectrum
        from matchms.exporting import save_as_msp

        # Create dummy spectrum
        spectrum = Spectrum(mz=numpy.array([100, 200, 300], dtype="float"),
                            intensities=numpy.array([10, 10, 500], dtype="float"),
                            metadata={"charge": -1,
                                      "inchi": '"InChI=1S/C6H12"',
                                      "precursor_mz": 222.2})

        # Write spectrum to test file
        save_as_msp(spectrum, "test.msp")

    Parameters
    ----------
    spectra:
        Expected input are match.Spectrum.Spectrum() objects.
    filename:
        Provide filename to save spectrum(s).
    write_peak_comments:
        Writes peak comments to individual peaks after the respective mz/intensity pair
        when set to True. Default is True.
    """
    file_extension = filename.split(".")[-1]
    assert file_extension.lower() not in _extentions_not_allowed, \
        f"File extension '.{file_extension}' not allowed."
    if not filename.endswith(".msp"):
        logger.warning("Spectra will be stored as msp file with extension .%s",
                       filename.split(".")[-1])
    spectra = _ensure_list(spectra)

    with open(filename, "w", encoding="utf-8") as outfile:
        for spectrum in spectra:
            _write_spectrum(spectrum, outfile, write_peak_comments)


def _write_spectrum(spectrum: Spectrum, outfile: IO, write_peak_comments: bool):
    _write_metadata(spectrum.metadata, outfile)
    if write_peak_comments is True:
        _write_peaks(spectrum.peaks, spectrum.peak_comments, outfile)
    else:
        _write_peaks(spectrum.peaks, None, outfile)
    outfile.write(os.linesep)


def _write_peaks(peaks: Spikes, peak_comments: Spectrum.peak_comments, outfile: IO):
    outfile.write(f"NUM PEAKS: {len(peaks)}\n")
    for mz, intensity in zip(peaks.mz, peaks.intensities):
        peak_comment = _format_peak_comment(mz, peak_comments)
        outfile.write(f"{mz}\t{intensity}{peak_comment}\n".expandtabs(12))


def _write_metadata(metadata: dict, outfile: IO):
    for key, value in metadata.items():
        if not (_is_num_peaks(key) or _is_peak_comments(key)):
            outfile.write(f"{key.upper()}: {value}\n")


def _format_peak_comment(mz: Union[int, float], peak_comments: Dict):
    """Format peak comment for given mz to return the quoted comment or empty string if no peak comment is present."""
    if peak_comments is None:
        return ""
    peak_comment = peak_comments.get(mz, None)
    if peak_comment is None:
        return ""
    return f"\t\"{peak_comment}\""


def _is_num_peaks(key: str) -> bool:
    return key.lower().startswith("num peaks")


def _is_peak_comments(key: str) -> bool:
    return key.lower().startswith("peak_comments")


def _ensure_list(spectra) -> List[Spectrum]:
    if not isinstance(spectra, list):
        # Assume that input was single Spectrum
        spectra = [spectra]
    return spectra



import numpy
from ..Spikes import Spikes
from ..typing import SpectrumType


def select_by_relative_intensity(spectrum_in: SpectrumType, intensity_from: float = 0.0,
                                 intensity_to: float = 1.0) -> SpectrumType:
    """Keep only peaks within set relative intensity range (keep if
    intensity_from >= intensity >= intensity_to).

    Parameters
    ----------
    intensity_from:
        Set lower threshold for relative peak intensity. Default is 0.0.
    intensity_to:
        Set upper threshold for relative peak intensity. Default is 1.0.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    assert intensity_from >= 0.0, "'intensity_from' should be larger than or equal to 0."
    assert intensity_to <= 1.0, "'intensity_to' should be smaller than or equal to 1.0."
    assert intensity_from <= intensity_to, "'intensity_from' should be smaller than or equal to 'intensity_to'."

    if len(spectrum.peaks) > 0:
        scale_factor = numpy.max(spectrum.peaks.intensities)
        normalized_intensities = spectrum.peaks.intensities / scale_factor
        condition = numpy.logical_and(intensity_from <= normalized_intensities, normalized_intensities <= intensity_to)
        spectrum.peaks = Spikes(mz=spectrum.peaks.mz[condition],
                                intensities=spectrum.peaks.intensities[condition])

    return spectrum



from .Scores import Scores
from .similarity.BaseSimilarity import BaseSimilarity
from .typing import QueriesType
from .typing import ReferencesType


def calculate_scores(references: ReferencesType, queries: QueriesType,
                     similarity_function: BaseSimilarity,
                     is_symmetric: bool = False) -> Scores:
    """Calculate the similarity between all reference objects versus all query objects.

    Example to calculate scores between 2 spectrums and iterate over the scores

    .. testcode::

        import numpy as np
        from matchms import calculate_scores, Spectrum
        from matchms.similarity import CosineGreedy

        spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.7, 0.2, 0.1]),
                              metadata={'id': 'spectrum1'})
        spectrum_2 = Spectrum(mz=np.array([100, 140, 190.]),
                              intensities=np.array([0.4, 0.2, 0.1]),
                              metadata={'id': 'spectrum2'})
        spectrums = [spectrum_1, spectrum_2]

        scores = calculate_scores(spectrums, spectrums, CosineGreedy())

        for (reference, query, score) in scores:
            print(f"Cosine score between {reference.get('id')} and {query.get('id')}" +
                  f" is {score['score']:.2f} with {score['matches']} matched peaks")

    Should output

    .. testoutput::

        Cosine score between spectrum1 and spectrum1 is 1.00 with 3 matched peaks
        Cosine score between spectrum1 and spectrum2 is 0.83 with 1 matched peaks
        Cosine score between spectrum2 and spectrum1 is 0.83 with 1 matched peaks
        Cosine score between spectrum2 and spectrum2 is 1.00 with 3 matched peaks

    Parameters
    ----------
    references
        List of reference objects
    queries
        List of query objects
    similarity_function
        Function which accepts a reference + query object and returns a score or tuple of scores
    is_symmetric
        Set to True when *references* and *queries* are identical (as for instance for an all-vs-all
        comparison). By using the fact that score[i,j] = score[j,i] the calculation will be about
        2x faster. Default is False.

    Returns
    -------

    ~matchms.Scores.Scores
    """

    return Scores(references=references, queries=queries,
                  similarity_function=similarity_function,
                  is_symmetric=is_symmetric).calculate()



"""
Functions for computing spectra similarities
############################################

Matchms provides a number of frequently used similarity scores to compare mass
spectra. This includes

* scores based on comparing peak positions and intensities
  (:class:`~matchms.similarity.CosineGreedy` or :class:`~matchms.similarity.ModifiedCosine`)
* simple scores that only assess precursor m/z or parent mass matches
  (:class:`~matchms.similarity.PrecursorMzMatch` or: :class:`~matchms.similarity.ParentMassMatch`)
* scores assessing molecular similarity if structures (SMILES, InchiKey) are given as metadata
  (:class:`~matchms.similarity.FingerprintSimilarity`)

It is also easily possible to add own custom similarity measures or import external ones
(such as `Spec2Vec <https://github.com/iomega/spec2vec>`_).
"""
from .CosineGreedy import CosineGreedy
from .CosineHungarian import CosineHungarian
from .FingerprintSimilarity import FingerprintSimilarity
from .IntersectMz import IntersectMz
from .ModifiedCosine import ModifiedCosine
from .ParentMassMatch import ParentMassMatch
from .PrecursorMzMatch import PrecursorMzMatch


__all__ = [
    "CosineGreedy",
    "CosineHungarian",
    "FingerprintSimilarity",
    "IntersectMz",
    "ModifiedCosine",
    "ParentMassMatch",
    "PrecursorMzMatch",
]



import logging
from ..typing import SpectrumType
from ..utils import convert_smiles_to_inchi
from ..utils import is_valid_inchi
from ..utils import is_valid_smiles


logger = logging.getLogger("matchms")


def derive_inchi_from_smiles(spectrum_in: SpectrumType) -> SpectrumType:
    """Find missing Inchi and derive from smiles where possible."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()
    inchi = spectrum.get("inchi")
    smiles = spectrum.get("smiles")

    if not is_valid_inchi(inchi) and is_valid_smiles(smiles):
        inchi = convert_smiles_to_inchi(smiles)
        if inchi:
            inchi = inchi.rstrip()
            spectrum.set("inchi", inchi)
            logger.info("Added InChI (%s) to metadata (was converted from smiles).", inchi)
        else:
            logger.warning("Could not convert smiles %s to InChI.", smiles)

    return spectrum



from typing import List
import numba
import numpy
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity


class ParentMassMatch(BaseSimilarity):
    """Return True if spectrums match in parent mass (within tolerance), and False otherwise.

    Example to calculate scores between 2 spectrums and iterate over the scores

    .. testcode::

        import numpy as np
        from matchms import calculate_scores
        from matchms import Spectrum
        from matchms.similarity import ParentMassMatch

        spectrum_1 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "1", "parent_mass": 100})
        spectrum_2 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "2", "parent_mass": 110})
        spectrum_3 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "3", "parent_mass": 103})
        spectrum_4 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "4", "parent_mass": 111})
        references = [spectrum_1, spectrum_2]
        queries = [spectrum_3, spectrum_4]

        similarity_score = ParentMassMatch(tolerance=5.0)
        scores = calculate_scores(references, queries, similarity_score)

        for (reference, query, score) in scores:
            print(f"Parentmass match between {reference.get('id')} and {query.get('id')}" +
                  f" is {score:.2f}")

    Should output

    .. testoutput::

        Parentmass match between 1 and 3 is 1.00
        Parentmass match between 1 and 4 is 0.00
        Parentmass match between 2 and 3 is 0.00
        Parentmass match between 2 and 4 is 1.00

    """
    # Set key characteristics as class attributes
    is_commutative = True
    # Set output data type, e.g.  "float" or [("score", "float"), ("matches", "int")]
    score_datatype = bool

    def __init__(self, tolerance: float = 0.1):
        """
        Parameters
        ----------
        tolerance
            Specify tolerance below which two masses are counted as match.
        """
        self.tolerance = tolerance

    def pair(self, reference: SpectrumType, query: SpectrumType) -> float:
        """Compare parent masses between reference and query spectrum.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.
        """
        parentmass_ref = reference.get("parent_mass")
        parentmass_query = query.get("parent_mass")
        assert parentmass_ref is not None and parentmass_query is not None, "Missing parent mass."

        score = abs(parentmass_ref - parentmass_query) <= self.tolerance
        return numpy.asarray(score, dtype=self.score_datatype)

    def matrix(self, references: List[SpectrumType], queries: List[SpectrumType],
               is_symmetric: bool = False) -> numpy.ndarray:
        """Compare parent masses between all references and queries.

        Parameters
        ----------
        references
            List/array of reference spectrums.
        queries
            List/array of Single query spectrums.
        is_symmetric
            Set to True when *references* and *queries* are identical (as for instance for an all-vs-all
            comparison). By using the fact that score[i,j] = score[j,i] the calculation will be about
            2x faster.
        """
        def collect_parentmasses(spectrums):
            """Collect parentmasses."""
            parentmasses = []
            for spectrum in spectrums:
                parentmass = spectrum.get("parent_mass")
                assert parentmass is not None, "Missing parent mass."
                parentmasses.append(parentmass)
            return numpy.asarray(parentmasses)

        parentmasses_ref = collect_parentmasses(references)
        parentmasses_query = collect_parentmasses(queries)
        if is_symmetric:
            return parentmass_scores_symmetric(parentmasses_ref, parentmasses_query,
                                               self.tolerance).astype(self.score_datatype)
        return parentmass_scores(parentmasses_ref, parentmasses_query,
                                 self.tolerance).astype(self.score_datatype)


@numba.njit
def parentmass_scores(parentmasses_ref, parentmasses_query, tolerance):
    scores = numpy.zeros((len(parentmasses_ref), len(parentmasses_query)))
    for i, parentmass_ref in enumerate(parentmasses_ref):
        for j, parentmass_query in enumerate(parentmasses_query):
            scores[i, j] = (abs(parentmass_ref - parentmass_query) <= tolerance)
    return scores


@numba.njit
def parentmass_scores_symmetric(parentmasses_ref, parentmasses_query, tolerance):
    scores = numpy.zeros((len(parentmasses_ref), len(parentmasses_query)))
    for i, parentmass_ref in enumerate(parentmasses_ref):
        for j in range(i, len(parentmasses_query)):
            scores[i, j] = (abs(parentmass_ref - parentmasses_query[j]) <= tolerance)
            scores[j, i] = scores[i, j]
    return scores



import logging
import re
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def derive_formula_from_name(spectrum_in: SpectrumType,
                             remove_formula_from_name: bool = True) -> SpectrumType:
    """Detect and remove misplaced formula in compound name and add to metadata.

    Method to find misplaced formulas in compound name based on regular expression.
    This will not chemically test the detected formula, so the search is limited
    to frequently occuring types of shape 'C47H83N1O8P1'.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    remove_formula_from_name:
        Remove found formula from compound name if set to True. Default is True.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    if spectrum.get("compound_name", None) is not None:
        name = spectrum.get("compound_name")
    else:
        assert spectrum.get("name", None) in [None, ""], ("Found 'name' but not 'compound_name' in metadata",
                                                          "Apply 'add_compound_name' filter first.")
        return spectrum

    # Detect formula at end of compound name
    end_of_name = name.split(" ")[-1]
    formula_from_name = end_of_name if looks_like_formula(end_of_name) else None

    if formula_from_name and remove_formula_from_name:
        name_formula_removed = " ".join(name.split(" ")[:-1])
        spectrum.set("compound_name", name_formula_removed)
        logger.info("Removed formula %s from compound name.", formula_from_name)

    # Add found formula to metadata (if not present yet)
    if formula_from_name and spectrum.get("formula", None) is None:
        spectrum.set("formula", formula_from_name)
        logger.info("Added formula %s to metadata.", formula_from_name)

    return spectrum


def looks_like_formula(formula):
    """Return True if input string has expected format of a molecular formula.
    Does only consider most frequent atoms found in many name strings.
    """
    regex_atoms = r"(C|F|H|N|O|P|S)"
    atom_count = len(re.findall(regex_atoms, formula))
    regexp = r"^(C|F|H|N|O|P|S|[0-9]|\(|\)){3,}$"
    return (atom_count > 2) and (re.search(regexp, formula) is not None)



from typing import Generator
from typing import TextIO
from typing import Union
import numpy
from pyteomics.mgf import MGF
from ..Spectrum import Spectrum


def load_from_mgf(source: Union[str, TextIO]) -> Generator[Spectrum, None, None]:
    """Load spectrum(s) from mgf file.

    This function will create ~matchms.Spectrum for every spectrum in the given
    .mgf file (or the file-like object).

    Examples:

    .. code-block:: python

        from matchms.importing import load_from_mgf

        file_mgf = "pesticides.mgf"
        spectra_from_path = list(load_from_mgf(file_mgf))

        # Or you can read the file in your application
        with open(file_mgf, 'r') as spectra_file:
            spectra_from_file = list(load_from_mgf(spectra_file))

    Parameters
    ----------
    source:
        Accepts both filename (with path) for .mgf file or a file-like
        object from a preloaded MGF file.
    """

    for pyteomics_spectrum in MGF(source, convert_arrays=1):

        metadata = pyteomics_spectrum.get("params", None)
        mz = pyteomics_spectrum["m/z array"]
        intensities = pyteomics_spectrum["intensity array"]

        # Sort by mz (if not sorted already)
        if not numpy.all(mz[:-1] <= mz[1:]):
            idx_sorted = numpy.argsort(mz)
            mz = mz[idx_sorted]
            intensities = intensities[idx_sorted]

        yield Spectrum(mz=mz, intensities=intensities, metadata=metadata)



from typing import List
import numba
import numpy
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity


class PrecursorMzMatch(BaseSimilarity):
    """Return True if spectrums match in precursor m/z (within tolerance), and False otherwise.
    The match within tolerance can be calculated based on an absolute m/z difference
    (tolerance_type="Dalton") or based on a relative difference in ppm (tolerance_type="ppm").

    Example to calculate scores between 2 pairs of spectrums and iterate over the scores

    .. testcode::

        import numpy as np
        from matchms import calculate_scores
        from matchms import Spectrum
        from matchms.similarity import PrecursorMzMatch

        spectrum_1 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "1", "precursor_mz": 100})
        spectrum_2 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "2", "precursor_mz": 110})
        spectrum_3 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "3", "precursor_mz": 103})
        spectrum_4 = Spectrum(mz=np.array([]),
                              intensities=np.array([]),
                              metadata={"id": "4", "precursor_mz": 111})
        references = [spectrum_1, spectrum_2]
        queries = [spectrum_3, spectrum_4]

        similarity_score = PrecursorMzMatch(tolerance=5.0, tolerance_type="Dalton")
        scores = calculate_scores(references, queries, similarity_score)

        for (reference, query, score) in scores:
            print(f"Precursor m/z match between {reference.get('id')} and {query.get('id')}" +
                  f" is {score:.2f}")

    Should output

    .. testoutput::

        Precursor m/z match between 1 and 3 is 1.00
        Precursor m/z match between 1 and 4 is 0.00
        Precursor m/z match between 2 and 3 is 0.00
        Precursor m/z match between 2 and 4 is 1.00

    """
    # Set key characteristics as class attributes
    is_commutative = True
    score_datatype = bool

    def __init__(self, tolerance: float = 0.1, tolerance_type: str = "Dalton"):
        """
        Parameters
        ----------
        tolerance
            Specify tolerance below which two m/z are counted as match.
        tolerance_type
            Chose between fixed tolerance in Dalton (="Dalton") or a relative difference
            in ppm (="ppm").
        """
        self.tolerance = tolerance
        assert tolerance_type in ["Dalton", "ppm"], "Expected type from ['Dalton', 'ppm']"
        self.type = tolerance_type

    def pair(self, reference: SpectrumType, query: SpectrumType) -> float:
        """Compare precursor m/z between reference and query spectrum.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.
        """
        precursormz_ref = reference.get("precursor_mz")
        precursormz_query = query.get("precursor_mz")
        assert precursormz_ref is not None and precursormz_query is not None, "Missing precursor m/z."

        if self.type == "Dalton":
            return abs(precursormz_ref - precursormz_query) <= self.tolerance

        mean_mz = (precursormz_ref + precursormz_query) / 2
        score = abs(precursormz_ref - precursormz_query)/mean_mz <= self.tolerance
        return numpy.asarray(score, dtype=self.score_datatype)

    def matrix(self, references: List[SpectrumType], queries: List[SpectrumType],
               is_symmetric: bool = False) -> numpy.ndarray:
        """Compare parent masses between all references and queries.

        Parameters
        ----------
        references
            List/array of reference spectrums.
        queries
            List/array of Single query spectrums.
        is_symmetric
            Set to True when *references* and *queries* are identical (as for instance for an all-vs-all
            comparison). By using the fact that score[i,j] = score[j,i] the calculation will be about
            2x faster.
        """
        def collect_precursormz(spectrums):
            """Collect precursors."""
            precursors = []
            for spectrum in spectrums:
                precursormz = spectrum.get("precursor_mz")
                assert precursormz is not None, "Missing precursor m/z."
                precursors.append(precursormz)
            return numpy.asarray(precursors)

        precursors_ref = collect_precursormz(references)
        precursors_query = collect_precursormz(queries)
        if is_symmetric and self.type == "Dalton":
            return precursormz_scores_symmetric(precursors_ref, precursors_query,
                                                self.tolerance).astype(self.score_datatype)
        if is_symmetric and self.type == "ppm":
            return precursormz_scores_symmetric_ppm(precursors_ref, precursors_query,
                                                    self.tolerance).astype(self.score_datatype)
        if self.type == "Dalton":
            return precursormz_scores(precursors_ref, precursors_query,
                                      self.tolerance).astype(self.score_datatype)
        return precursormz_scores_ppm(precursors_ref, precursors_query,
                                      self.tolerance).astype(self.score_datatype)


@numba.njit
def precursormz_scores(precursors_ref, precursors_query, tolerance):
    scores = numpy.zeros((len(precursors_ref), len(precursors_query)))
    for i, precursormz_ref in enumerate(precursors_ref):
        for j, precursormz_query in enumerate(precursors_query):
            scores[i, j] = (abs(precursormz_ref - precursormz_query) <= tolerance)
    return scores


@numba.njit
def precursormz_scores_symmetric(precursors_ref, precursors_query, tolerance):
    scores = numpy.zeros((len(precursors_ref), len(precursors_query)))
    for i, precursormz_ref in enumerate(precursors_ref):
        for j in range(i, len(precursors_query)):
            scores[i, j] = (abs(precursormz_ref - precursors_query[j]) <= tolerance)
            scores[j, i] = scores[i, j]
    return scores


@numba.njit
def precursormz_scores_ppm(precursors_ref, precursors_query, tolerance_ppm):
    scores = numpy.zeros((len(precursors_ref), len(precursors_query)))
    for i, precursormz_ref in enumerate(precursors_ref):
        for j, precursormz_query in enumerate(precursors_query):
            mean_mz = (precursormz_ref + precursormz_query)/2
            scores[i, j] = (abs(precursormz_ref - precursormz_query)/mean_mz * 1e6 <= tolerance_ppm)
    return scores


@numba.njit
def precursormz_scores_symmetric_ppm(precursors_ref, precursors_query, tolerance_ppm):
    scores = numpy.zeros((len(precursors_ref), len(precursors_query)))
    for i, precursormz_ref in enumerate(precursors_ref):
        for j in range(i, len(precursors_query)):
            mean_mz = (precursormz_ref + precursors_query[j])/2
            diff_ppm = abs(precursormz_ref - precursors_query[j])/mean_mz * 1e6
            scores[i, j] = (diff_ppm <= tolerance_ppm)
            scores[j, i] = scores[i, j]
    return scores



from __future__ import annotations
import numpy
from deprecated.sphinx import deprecated
from matchms.similarity.BaseSimilarity import BaseSimilarity
from matchms.typing import QueriesType
from matchms.typing import ReferencesType


class Scores:
    """Contains reference and query spectrums and the scores between them.

    The scores can be retrieved as a matrix with the :py:attr:`Scores.scores` attribute.
    The reference spectrum, query spectrum, score pairs can also be iterated over in query then reference order.

    Example to calculate scores between 2 spectrums and iterate over the scores

    .. testcode::

        import numpy as np
        from matchms import calculate_scores
        from matchms import Spectrum
        from matchms.similarity import CosineGreedy

        spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.7, 0.2, 0.1]),
                              metadata={'id': 'spectrum1'})
        spectrum_2 = Spectrum(mz=np.array([100, 140, 190.]),
                              intensities=np.array([0.4, 0.2, 0.1]),
                              metadata={'id': 'spectrum2'})
        spectrum_3 = Spectrum(mz=np.array([110, 140, 195.]),
                              intensities=np.array([0.6, 0.2, 0.1]),
                              metadata={'id': 'spectrum3'})
        spectrum_4 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.6, 0.1, 0.6]),
                              metadata={'id': 'spectrum4'})
        references = [spectrum_1, spectrum_2]
        queries = [spectrum_3, spectrum_4]

        similarity_measure = CosineGreedy()
        scores = calculate_scores(references, queries, similarity_measure)

        for (reference, query, score) in scores:
            print(f"Cosine score between {reference.get('id')} and {query.get('id')}" +
                  f" is {score['score']:.2f} with {score['matches']} matched peaks")

    Should output

    .. testoutput::

        Cosine score between spectrum1 and spectrum3 is 0.00 with 0 matched peaks
        Cosine score between spectrum1 and spectrum4 is 0.80 with 3 matched peaks
        Cosine score between spectrum2 and spectrum3 is 0.14 with 1 matched peaks
        Cosine score between spectrum2 and spectrum4 is 0.61 with 1 matched peaks
    """
    def __init__(self, references: ReferencesType, queries: QueriesType,
                 similarity_function: BaseSimilarity, is_symmetric: bool = False):
        """

        Parameters
        ----------
        references
            List of reference objects
        queries
            List of query objects
        similarity_function
            Expected input is an object based on :class:`~matchms.similarity.BaseSimilarity`. It is
            expected to provide a *.pair()* and *.matrix()* method for computing similarity scores between
            references and queries.
        is_symmetric
            Set to True when *references* and *queries* are identical (as for instance for an all-vs-all
            comparison). By using the fact that score[i,j] = score[j,i] the calculation will be about
            2x faster. Default is False.
        """
        Scores._validate_input_arguments(references, queries, similarity_function)

        self.n_rows = len(references)
        self.n_cols = len(queries)
        self.references = numpy.asarray(references)
        self.queries = numpy.asarray(queries)
        self.similarity_function = similarity_function
        self.is_symmetric = is_symmetric
        self._scores = numpy.empty([self.n_rows, self.n_cols], dtype="object")
        self._index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < self.scores.size:
            # pylint: disable=unbalanced-tuple-unpacking
            r, c = numpy.unravel_index(self._index, self._scores.shape)
            self._index += 1
            result = self._scores[r, c]
            if not isinstance(result, tuple):
                result = (result,)
            return (self.references[r], self.queries[c]) + result
        self._index = 0
        raise StopIteration

    def __str__(self):
        return self._scores.__str__()

    @staticmethod
    def _validate_input_arguments(references, queries, similarity_function):
        assert isinstance(references, (list, tuple, numpy.ndarray)),\
            "Expected input argument 'references' to be list or tuple or numpy.ndarray."

        assert isinstance(queries, (list, tuple, numpy.ndarray)),\
            "Expected input argument 'queries' to be list or tuple or numpy.ndarray."

        assert isinstance(similarity_function, BaseSimilarity),\
            "Expected input argument 'similarity_function' to have BaseSimilarity as super-class."

    @deprecated(version='0.6.0', reason="Calculate scores via calculate_scores() function.")
    def calculate(self) -> Scores:
        """
        Calculate the similarity between all reference objects v all query objects using
        the most suitable available implementation of the given similarity_function.
        Advised method to calculate similarity scores is :meth:`~matchms.calculate_scores`.
        """
        if self.n_rows == self.n_cols == 1:
            self._scores[0, 0] = self.similarity_function.pair(self.references[0],
                                                               self.queries[0])
        else:
            self._scores = self.similarity_function.matrix(self.references,
                                                           self.queries,
                                                           is_symmetric=self.is_symmetric)
        return self

    def scores_by_reference(self, reference: ReferencesType,
                            sort: bool = False) -> numpy.ndarray:
        """Return all scores for the given reference spectrum.

        Parameters
        ----------
        reference
            Single reference Spectrum.
        sort
            Set to True to obtain the scores in a sorted way (relying on the
            :meth:`~.BaseSimilarity.sort` function from the given similarity_function).
        """
        assert reference in self.references, "Given input not found in references."
        selected_idx = int(numpy.where(self.references == reference)[0])
        if sort:
            query_idx_sorted = self.similarity_function.sort(self._scores[selected_idx, :])
            return list(zip(self.queries[query_idx_sorted],
                            self._scores[selected_idx, query_idx_sorted].copy()))
        return list(zip(self.queries, self._scores[selected_idx, :].copy()))

    def scores_by_query(self, query: QueriesType, sort: bool = False) -> numpy.ndarray:
        """Return all scores for the given query spectrum.

        For example

        .. testcode::

            import numpy as np
            from matchms import calculate_scores, Scores, Spectrum
            from matchms.similarity import CosineGreedy

            spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                                  intensities=np.array([0.7, 0.2, 0.1]),
                                  metadata={'id': 'spectrum1'})
            spectrum_2 = Spectrum(mz=np.array([100, 140, 190.]),
                                  intensities=np.array([0.4, 0.2, 0.1]),
                                  metadata={'id': 'spectrum2'})
            spectrum_3 = Spectrum(mz=np.array([110, 140, 195.]),
                                  intensities=np.array([0.6, 0.2, 0.1]),
                                  metadata={'id': 'spectrum3'})
            spectrum_4 = Spectrum(mz=np.array([100, 150, 200.]),
                                  intensities=np.array([0.6, 0.1, 0.6]),
                                  metadata={'id': 'spectrum4'})
            references = [spectrum_1, spectrum_2, spectrum_3]
            queries = [spectrum_2, spectrum_3, spectrum_4]

            scores = calculate_scores(references, queries, CosineGreedy())
            selected_scores = scores.scores_by_query(spectrum_4, sort=True)
            print([x[1]["score"].round(3) for x in selected_scores])

        Should output

        .. testoutput::

            [0.796, 0.613, 0.0]

        Parameters
        ----------
        query
            Single query Spectrum.
        sort
            Set to True to obtain the scores in a sorted way (relying on the
            :meth:`~.BaseSimilarity.sort` function from the given similarity_function).

        """
        assert query in self.queries, "Given input not found in queries."
        selected_idx = int(numpy.where(self.queries == query)[0])
        if sort:
            references_idx_sorted = self.similarity_function.sort(self._scores[:, selected_idx])
            return list(zip(self.references[references_idx_sorted],
                            self._scores[references_idx_sorted, selected_idx].copy()))
        return list(zip(self.references, self._scores[:, selected_idx].copy()))

    @property
    def scores(self) -> numpy.ndarray:
        """Scores as numpy array

        For example

        .. testcode::

            import numpy as np
            from matchms import calculate_scores, Scores, Spectrum
            from matchms.similarity import IntersectMz

            spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                                  intensities=np.array([0.7, 0.2, 0.1]))
            spectrum_2 = Spectrum(mz=np.array([100, 140, 190.]),
                                  intensities=np.array([0.4, 0.2, 0.1]))
            spectrums = [spectrum_1, spectrum_2]

            scores = calculate_scores(spectrums, spectrums, IntersectMz()).scores

            print(scores[0].dtype)
            print(scores.shape)
            print(scores)

        Should output

        .. testoutput::

             float64
             (2, 2)
             [[1.  0.2]
              [0.2 1. ]]
        """
        return self._scores.copy()



import re
from typing import Generator
from typing import Iterator
from typing import Tuple
import numpy
from ..Spectrum import Spectrum


def parse_msp_file(filename: str) -> Generator[dict, None, None]:
    """Read msp file and parse info in list of spectrum dictionaries."""

    # Lists/dicts that will contain all params, masses and intensities of each molecule
    params = {}
    masses = []
    intensities = []
    peak_comments = {}

    # Peaks counter. Used to track and count the number of peaks
    peakscount = 0

    with open(filename, 'r', encoding='utf-8') as f:
        for line in f:
            rline = line.rstrip()

            if len(rline) == 0:
                continue

            if contains_metadata(rline):
                parse_metadata(rline, params)
            else:
                # Obtaining the masses and intensities
                peak_pairs = get_peak_tuples(rline)

                for peak in peak_pairs:
                    mz, intensity = get_peak_values(peak)
                    comment = get_peak_comment(peak)
                    if comment is not None:
                        peak_comments.update({mz: comment})

                    peakscount += 1
                    masses.append(mz)
                    intensities.append(intensity)

                # Obtaining the masses and intensities
                if int(params['num peaks']) == peakscount:
                    peakscount = 0
                    yield {
                        'params': (params),
                        'm/z array': numpy.array(masses),
                        'intensity array': numpy.array(intensities),
                        'peak comments': peak_comments
                    }

                    params = {}
                    masses = []
                    intensities = []
                    peak_comments = {}


def get_peak_values(peak: str) -> Tuple[float, float]:
    """ Get the m/z and intensity value from the line containing the peak information. """
    splitted_line = peak.split(maxsplit=2)
    mz = float(splitted_line[0].strip())
    intensity = float(splitted_line[1].strip())
    return mz, intensity


def get_peak_tuples(rline: str) -> Iterator[str]:
    """ Splits line at ';' and performs additional string cleaning. """
    tokens = filter(None, rline.split(";"))
    peak_pairs = map(lambda x: x.lstrip().rstrip(), tokens)
    return peak_pairs


def get_peak_comment(rline: str) -> str:
    """ Get the peak comment from the line containing the peak information. """
    try:
        comment = re.findall(r'[\"\'](.*)[\"\']', rline)[0]
    except IndexError:
        comment = None
    return comment


def parse_metadata(rline: str, params: dict):
    """ Reads metadata contained in line into params dict. """
    splitted_line = rline.split(":", 1)
    if splitted_line[0].lower() == 'comments':
        # Obtaining the parameters inside the comments index
        for s in splitted_line[1][2:-1].split('" "'):
            splitted_line = s.split("=", 1)
            if splitted_line[0].lower() in params.keys() and splitted_line[0].lower() == 'smiles':
                params[splitted_line[0].lower()+"_2"] = splitted_line[1].strip()
            else:
                params[splitted_line[0].lower()] = splitted_line[1].strip()
    else:
        params[splitted_line[0].lower()] = splitted_line[1].strip()


def contains_metadata(rline: str) -> bool:
    """ Check if line contains Spectrum metadata."""
    return ':' in rline


def load_from_msp(filename: str) -> Generator[Spectrum, None, None]:
    """
    MSP file to a :py:class:`~matchms.Spectrum.Spectrum` objects
    Function that reads a .msp file and converts the info
    in :py:class:`~matchms.Spectrum.Spectrum` objects.

    Parameters
    ----------
    filename:
        Path of the msp file.

    Yields
    ------
    Yield a spectrum object with the data of the msp file


    Example:

    .. code-block:: python

        from matchms.importing import load_from_msp

        # Download msp file from MassBank of North America repository at https://mona.fiehnlab.ucdavis.edu/
        file_msp = "MoNA-export-GC-MS-first10.msp"
        spectrums = list(load_from_msp(file_msp))
    """

    for spectrum in parse_msp_file(filename):
        metadata = spectrum.get("params", None)
        mz = spectrum["m/z array"]
        intensities = spectrum["intensity array"]
        peak_comments = spectrum["peak comments"]
        if peak_comments != {}:
            metadata["peak_comments"] = peak_comments

        # Sort by mz (if not sorted already)
        if not numpy.all(mz[:-1] <= mz[1:]):
            idx_sorted = numpy.argsort(mz)
            mz = mz[idx_sorted]
            intensities = intensities[idx_sorted]

        yield Spectrum(mz=mz, intensities=intensities, metadata=metadata)



from typing import Optional
from typing import Union
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np


_annotation_kws = {
    "horizontalalignment": "left",  # if not mirror_intensity else "right",
    "verticalalignment": "center",
    "fontsize": 7,
    "rotation": 90,
    "rotation_mode": "anchor",
    "zorder": 5,
}


def plot_spectrum(spectrum,
                  annotate_ions: bool = False,
                  mirror_intensity: bool = False,
                  grid: Union[bool, str] = True,
                  ax: plt.Axes = None,
                  peak_color="teal",
                  **plt_kwargs) -> plt.Axes:
    """
    Plot a single MS/MS spectrum.

    Code is largely taken from package "spectrum_utils".

    Parameters
    ----------
    spectrum: matchms.Spectrum
        The spectrum to be plotted.
    annotate_ions:
        Flag indicating whether or not to annotate fragment using peak comments
        (if present in the spectrum). The default is True.
    mirror_intensity:
        Flag indicating whether to flip the intensity axis or not.
    grid:
        Draw grid lines or not. Either a boolean to enable/disable both major
        and minor grid lines or 'major'/'minor' to enable major or minor grid
        lines respectively.
    ax:
        Axes instance on which to plot the spectrum. If None the current Axes
        instance is used.

    Returns
    -------
    plt.Axes
        The matplotlib Axes instance on which the spectrum is plotted.
    """
    # pylint: disable=too-many-locals, too-many-arguments
    if ax is None:
        ax = plt.gca()

    min_mz = max(0, np.floor(spectrum.peaks.mz[0] / 100 - 1) * 100)
    max_mz = np.ceil(spectrum.peaks.mz[-1] / 100 + 1) * 100
    max_intensity = spectrum.peaks.intensities.max()

    intensities = spectrum.peaks.intensities / max_intensity

    def make_stems():
        """calculate where the stems of the spectrum peaks are going to be"""
        x = np.zeros([2, spectrum.peaks.mz.size], dtype="float")
        y = np.zeros(x.shape)
        x[:, :] = np.tile(spectrum.peaks.mz, (2, 1))
        y[1, :] = intensities
        return x, y

    x, y = make_stems()
    if mirror_intensity is True:
        y = -y
    ax.plot(x, y, color=peak_color, linewidth=1.0, marker="", zorder=5, **plt_kwargs)
    if annotate_ions and isinstance(spectrum.get("peak_comments"), dict):
        for mz, comment in spectrum.get("peak_comments").items():
            idx = (-abs(spectrum.peaks.mz - mz)).argmax()
            ax.text(mz, intensities[idx], f"m/z: {mz} \n {comment}",
                    _annotation_kws)

    ax.set_xlim(min_mz, max_mz)
    ax.yaxis.set_major_formatter(mticker.PercentFormatter(xmax=1.0))
    y_max = 1.25 if annotate_ions else 1.10
    ax.set_ylim(*(0, y_max) if not mirror_intensity else (-y_max, 0))

    ax.xaxis.set_minor_locator(mticker.AutoLocator())
    ax.yaxis.set_minor_locator(mticker.AutoLocator())
    ax.xaxis.set_minor_locator(mticker.AutoMinorLocator())
    ax.yaxis.set_minor_locator(mticker.AutoMinorLocator())
    if grid in (True, "both", "major"):
        ax.grid(visible=True, which="major", color="#9E9E9E", linewidth=0.2)
    if grid in (True, "both", "minor"):
        ax.grid(visible=True, which="minor", color="#9E9E9E", linewidth=0.2)
    ax.set_axisbelow(True)

    ax.tick_params(axis="both", which="both", labelsize="small")
    y_ticks = ax.get_yticks()
    ax.set_yticks(y_ticks[y_ticks <= 1.0])

    ax.set_xlabel("m/z", style="italic")
    ax.set_ylabel("Intensity")
    title = "Spectrum" if spectrum.get("compound_name") is None else spectrum.get("compound_name")
    ax.set_title(title)
    return ax


def plot_spectra_mirror(spec_top,
                        spec_bottom,
                        ax: Optional[plt.Axes] = None,
                        **spectrum_kws) -> plt.Axes:
    """Mirror plot two MS/MS spectra.

    Code is largely taken from package "spectrum_utils".

    Parameters
    ----------
    spec_top: matchms.Spectrum
        The spectrum to be plotted on the top.
    spec_bottom: matchms.Spectrum
        The spectrum to be plotted on the bottom.
    ax:
        Axes instance on which to plot the spectrum. If None the current Axes
        instance is used.
    spectrum_kws:
        Keyword arguments for `plot_spectrum()`.

    Returns
    -------
    plt.Axes
        The matplotlib Axes instance on which the spectra are plotted.
    """
    if ax is None:
        ax = plt.gca()

    if spectrum_kws is None:
        spectrum_kws = {}
    # Top spectrum.
    plot_spectrum(spec_top, mirror_intensity=False, ax=ax, peak_color="darkblue", **spectrum_kws)
    y_max = ax.get_ylim()[1]

    # Mirrored bottom spectrum.
    plot_spectrum(spec_bottom, mirror_intensity=True, ax=ax, peak_color="teal", **spectrum_kws)
    y_min = ax.get_ylim()[0]
    ax.set_ylim(y_min, y_max)

    ax.axhline(0, color="#9E9E9E", zorder=10)

    # Update axes so that both spectra fit.
    min_mz = max(
        [
            0,
            np.floor(spec_top.peaks.mz[0] / 100 - 1) * 100,
            np.floor(spec_bottom.peaks.mz[0] / 100 - 1) * 100,
        ]
    )
    max_mz = max(
        [
            np.ceil(spec_top.peaks.mz[-1] / 100 + 1) * 100,
            np.ceil(spec_bottom.peaks.mz[-1] / 100 + 1) * 100,
        ]
    )
    ax.set_xlim(min_mz, max_mz)
    ax.yaxis.set_major_locator(mticker.AutoLocator())
    ax.yaxis.set_minor_locator(mticker.AutoMinorLocator())
    ax.yaxis.set_major_formatter(
        mticker.FuncFormatter(lambda x, pos: f"{abs(x):.0%}")
    )

    name1 = "Spectrum 1" if spec_top.get("compound_name") is None else spec_top.get("compound_name")
    name2 = "Spectrum 2" if spec_bottom.get("compound_name") is None else spec_bottom.get("compound_name")

    x_text = 0.04 * (max_mz - min_mz)
    ax.text(x_text, y_max, name1, ha="left", va="top", zorder=2, backgroundcolor="white")
    ax.text(x_text, y_min, name2, ha="left", va="bottom", zorder=2, backgroundcolor="white")
    ax.set_title("Spectrum comparison")
    return ax


def plot_spectra_array(spectrums,
                       n_cols: int = 2,
                       peak_color="darkblue",
                       dpi: int = 200,
                       **spectrum_kws) -> plt.Axes:
    """Mirror plot two MS/MS spectra.

    Code is largely taken from package "spectrum_utils".

    Parameters
    ----------
    spectrums: list of matchms.Spectrum
        List of spectra to be plotted in a single figure.
    n_cols:
        Number of spectra to be plotted per row. Default is 4.
    spectrum_kws:
        Keyword arguments for `plot_spectrum()`.

    """
    assert isinstance(spectrums, list), "Expected list of Spectrum objects as input."
    n_spectra = len(spectrums)
    n_rows = int(np.ceil(n_spectra / n_cols))
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(7 * n_cols, 3 * n_rows), dpi=dpi)

    if spectrum_kws is None:
        spectrum_kws = {}

    for i in range(n_rows):
        for j in range(n_cols):
            counter = i * n_cols + j
            if counter >= n_spectra:
                break

            plot_spectrum(spectrums[counter],
                          mirror_intensity=False, ax=axes[i, j],
                          peak_color=peak_color, **spectrum_kws)
            axes[i, j].set_title("")
            if spectrums[counter].get("compound_name") is None:
                name = f"Spectrum {i * n_cols + j}"
            else:
                name = spectrums[counter].get("compound_name")

            y_max = axes[i, j].get_ylim()[1]
            x_min = axes[i, j].get_xlim()[0]
            axes[i, j].text(x_min, y_max, name, va="bottom", zorder=2)

    plt.title("Spectrum comparison")
    return fig, axes



from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
import numpy
from .Spectrum import Spectrum


SpectrumType = Optional[Spectrum]
ReferencesType = QueriesType = Union[List[object], Tuple[object], numpy.ndarray]

"""Input for a similarity function"""
Sample = Union[Spectrum, None, object]

"""Result of a similarity function"""
Score = Union[float, Tuple[float, int]]



import numpy
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity


class IntersectMz(BaseSimilarity):
    """Example score for illustrating how to build custom spectra similarity score.

    IntersectMz will count all exact matches of peaks and divide it by all unique
    peaks found in both spectrums.

    Example of how matchms similarity functions can be used:

    .. testcode::

        import numpy as np
        from matchms import Spectrum
        from matchms.similarity import IntersectMz

        spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.7, 0.2, 0.1]))
        spectrum_2 = Spectrum(mz=np.array([100, 140, 190.]),
                              intensities=np.array([0.4, 0.2, 0.1]))

        # Construct a similarity function
        similarity_measure = IntersectMz(scaling=1.0)

        score = similarity_measure.pair(spectrum_1, spectrum_2)

        print(f"IntersectMz score is {score:.2f}")

    Should output

    .. testoutput::

        IntersectMz score is 0.20

    """

    def __init__(self, scaling: float = 1.0):
        """Constructor. Here, function parameters are defined.

        Parameters
        ----------
        scaling
            Scale scores to maximum possible score being 'scaling'.
        """
        self.scaling = scaling

    def pair(self, reference: SpectrumType, query: SpectrumType) -> float:
        """This will calculate the similarity score between two spectra."""
        mz_ref = set(reference.peaks.mz)
        mz_query = set(query.peaks.mz)
        intersected = mz_query.intersection(mz_ref)
        unioned = mz_query.union(mz_ref)

        if len(unioned) == 0:
            return 0

        return numpy.float64(self.scaling * len(intersected) / len(unioned))



import ast
import json
import logging
from typing import List
from typing import Union
import numpy
from ..Spectrum import Spectrum


logger = logging.getLogger("matchms")


def load_from_json(filename: str) -> List[Spectrum]:
    """Load spectrum(s) from json file.

    JSON document formatted like the `GNPS Spectra library <https://gnps-external.ucsd.edu/gnpslibrary>`_.
    Spectrums with zero peaks will be skipped.

    Example:

    .. code-block:: python

        from matchms.importing import load_from_json

        file_json = "gnps_testdata.json"
        spectrums = load_from_json(file_json)

    Parameters
    ----------
    filename
        Provide filename for json file containing spectrum(s).

    """
    with open(filename, 'rb') as fin:
        spectrums = []
        for spectrum_dict in json.load(fin):
            spectrum = as_spectrum(spectrum_dict)
            if spectrum is not None:
                spectrums.append(spectrum)

    return spectrums


def as_spectrum(dct: dict) -> Union[dict, Spectrum, None]:
    """A :py:func:`json.load` object_hook to convert dictionary shaped like
    spectrum into :py:class:`~matchms.Spectrum.Spectrum` object.

    Parameters
    ----------
    dct
        Dictionary shaped like spectrum
    Returns
    -------
    A Spectrum or None when no peaks where found.
    """
    # Recognize Spectrum by peaks_json key
    if 'peaks_json' in dct:
        return dict2spectrum(dct)
    return None


def dict2spectrum(spectrum_dict: dict) -> Union[Spectrum, None]:
    """Convert dictionary to a :py:class:`~matchms.Spectrum.Spectrum` object.

    Parameters
    ----------
    spectrum_dict
        Dictionary shaped like a single JSON object from the `GNPS Spectra library <https://gnps-external.ucsd.edu/gnpslibrary>`_

    Returns
    -------
    A Spectrum or None when no peaks where found.
    """
    not_metadata_fields = ["peaks_json"]
    parse_fieldnames = dict(inchi_aux="inchiaux",
                            ion_mode="ionmode")

    def get_peaks_list(spectrum_dict, fieldname):
        peaks_list = spectrum_dict.get(fieldname)
        if isinstance(peaks_list, list):
            return peaks_list
        # Handle peaks list when stored as string
        if isinstance(peaks_list, str):
            return ast.literal_eval(peaks_list)
        return []

    def parse_fieldname(key):
        """Add options to read GNPS style json files."""
        key_parsed = key.lower()
        key_parsed = parse_fieldnames.get(key_parsed, key_parsed)
        return key_parsed

    metadata_dict = {parse_fieldname(key): spectrum_dict[key]
                     for key in spectrum_dict if key not in not_metadata_fields}
    peaks_list = get_peaks_list(spectrum_dict, "peaks_json")
    if len(peaks_list) > 0 and metadata_dict:
        mz = numpy.array(peaks_list)[:, 0]
        intensities = numpy.array(peaks_list)[:, 1]

        # Sort by mz (if not sorted already)
        if not numpy.all(mz[:-1] <= mz[1:]):
            idx_sorted = numpy.argsort(mz)
            mz = mz[idx_sorted]
            intensities = intensities[idx_sorted]
        return Spectrum(mz=mz,
                        intensities=intensities,
                        metadata=metadata_dict)
    logger.info("Empty spectrum found (no peaks in 'peaks_json'). Will not be imported.")
    return None



import numpy
from ..Spikes import Spikes
from ..typing import SpectrumType


def select_by_mz(spectrum_in: SpectrumType, mz_from: float = 0.0,
                 mz_to: float = 1000.0) -> SpectrumType:
    """Keep only peaks between mz_from and mz_to (keep if mz_from >= m/z >= mz_to).

    Parameters
    ----------
    mz_from:
        Set lower threshold for m/z peak positions. Default is 0.0.
    mz_to:
        Set upper threshold for m/z peak positions. Default is 1000.0.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    assert mz_from <= mz_to, "'mz_from' should be smaller than or equal to 'mz_to'."

    condition = numpy.logical_and(mz_from <= spectrum.peaks.mz, spectrum.peaks.mz <= mz_to)

    spectrum.peaks = Spikes(mz=spectrum.peaks.mz[condition],
                            intensities=spectrum.peaks.intensities[condition])

    return spectrum



import csv
import os
from functools import lru_cache
from typing import Dict
import numpy


@lru_cache(maxsize=4)
def load_adducts_dict() -> Dict[str, dict]:
    """Load dictionary of known adducts containing the adduct mass and charge.
    Makes sure that file loading is cached.

    Adduct information is based on information from
    https://fiehnlab.ucdavis.edu/staff/kind/metabolomics/ms-adduct-calculator/
    and was extended by F.Huber and JJJ.v.d.Hooft.

    The full table can be found at
    https://github.com/matchms/matchms/blob/expand_adducts/matchms/data/known_adducts_table.csv

    TODO: change to relative path link or update link

    """
    known_adducts_file = os.path.join(os.path.dirname(__file__), "..", "data", "known_adducts_table.csv")
    assert os.path.isfile(known_adducts_file), "Could not find known_adducts_table.csv."

    with open(known_adducts_file, newline='', encoding='utf-8-sig') as csvfile:
        reader = csv.DictReader(csvfile)
        adducts_dict = {}
        for row in reader:
            assert "adduct" in row
            adducts_dict[row["adduct"]] = {x[0]: x[1] for x in row.items() if x[0] != "adduct"}

    return _convert_and_fill_dict(adducts_dict)


@lru_cache(maxsize=4)
def load_known_adduct_conversions() -> Dict[str, dict]:
    """Load dictionary of known adduct conversions. Makes sure that file loading is cached.
    """
    adduct_conversions_file = os.path.join(os.path.dirname(__file__), "..", "data", "known_adduct_conversions.csv")
    assert os.path.isfile(adduct_conversions_file), "Could not find known_adduct_conversions.csv."

    with open(adduct_conversions_file, newline='', encoding='utf-8-sig') as csvfile:
        reader = csv.DictReader(csvfile)
        known_adduct_conversions = {}
        for row in reader:
            known_adduct_conversions[row['input_adduct']] = row['corrected_adduct']

    return known_adduct_conversions


def _convert_and_fill_dict(adduct_dict: Dict[str, dict]) -> Dict[str, dict]:
    """Convert string entries to int/float and fill missing entries ('n/a')
    with best basic guesses."""
    def is_float(value):
        try:
            float(value)
            return True
        except ValueError:
            return False

    def is_int(value):
        try:
            int(value)
            return True
        except ValueError:
            return False

    filled_dict = {}
    for adduct, values in adduct_dict.items():
        ionmode = values["ionmode"]
        charge = values["charge"]
        mass_multiplier = values["mass_multiplier"]
        correction_mass = values["correction_mass"]

        if is_int(charge):
            charge = int(charge)
        else:
            charge = 1 * (ionmode == "positive") - 1 * (ionmode == "negative")
        values["charge"] = charge

        if is_float(mass_multiplier):
            mass_multiplier = float(mass_multiplier)
        else:
            mass_multiplier = 1.0
        values["mass_multiplier"] = mass_multiplier

        if is_float(correction_mass):
            correction_mass = float(correction_mass)
        else:
            correction_mass = 1.007276 * numpy.sign(charge)
        values["correction_mass"] = correction_mass

        filled_dict[adduct] = values
    return filled_dict



"""Matchms logger.

Matchms functions and method report unexpected or undesired behavior as
logging WARNING, and additional information as INFO.
The default logging level is set to WARNING. If you want to output additional
logging messages, you can lower the logging level to INFO using set_matchms_logger_level:

.. code-block:: python

    from matchms import set_matchms_logger_level

    set_matchms_logger_level("INFO")

If you want to suppress logging warnings, you can also raise the logging level
to ERROR by:

.. code-block:: python

    set_matchms_logger_level("ERROR")

To write logging entries to a local file, you can do the following:

.. code-block:: python

    from matchms.logging_functions import add_logging_to_file

    add_logging_to_file("sample.log", loglevel="INFO")

If you want to write the logging messages to a local file while silencing the
stream of such messages, you can do the following:

.. code-block:: python

    from matchms.logging_functions import add_logging_to_file

    add_logging_to_file("sample.log", loglevel="INFO",
                        remove_stream_handlers=True)

"""
import logging
import logging.config
import sys


_formatter = logging.Formatter(
    '%(asctime)s:%(levelname)s:%(name)s:%(module)s:%(message)s')


def _init_logger(logger_name="matchms"):
    """Initialize matchms logger."""
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.WARNING)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.WARNING)
    handler.setFormatter(_formatter)
    logger.addHandler(handler)
    logger.info('Completed configuring matchms logger.')


def set_matchms_logger_level(loglevel: str, logger_name="matchms"):
    """Update logging level to given loglevel.

    Parameters
    ----------
    loglevels
        Can be 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'.
    logger_name
        Default is "matchms". Change if logger name should be different.
    """
    level = logging.getLevelName(loglevel)
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)


def add_logging_to_file(filename: str, loglevel: str = "INFO",
                        remove_stream_handlers: bool = False,
                        logger_name="matchms"):
    """Add logging to file.

    Current implementation does not change the initial logging stream,
    but simply adds a FileHandler to write logging entries to a file.

    Parameters
    ----------
    filename
        Name of file for write logging output.
    loglevels
        Can be 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'.
    remove_stream_handlers
        Set to True if only logging to file is desired.
    logger_name
        Default is "matchms". Change if logger name should be different.
    """
    level = logging.getLevelName(loglevel)
    logger = logging.getLogger(logger_name)
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(_formatter)
    logger.addHandler(file_handler)

    # Remove the existing handlers if they are not of type FileHandler
    if remove_stream_handlers is True:
        for handler in logger.handlers:
            if not isinstance(handler, logging.FileHandler):
                logger.removeHandler(handler)


def reset_matchms_logger(logger_name="matchms"):
    """Reset matchms logger to initial state.

    This will remove all logging Handlers and initialize a new matchms logger.
    Use this function to reset previous changes made to the default matchms logger.

    Parameters
    ----------
    logger_name
        Default is "matchms". Change if logger name should be different.
    """
    logger = logging.getLogger(logger_name)
    logger.handlers.clear()
    _init_logger()



import logging
from typing import Any
from typing import List
from typing import Optional
from matchms.utils import filter_none
from matchms.utils import get_common_keys
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


_retention_time_keys = ["retention_time", "retentiontime", "rt", "scan_start_time", "RT_Query"]
_retention_index_keys = ["retention_index", "retentionindex", "ri"]


def safe_store_value(spectrum: SpectrumType, value: Any, target_key: str) -> SpectrumType:
    """Helper function to safely store a value in the target key without throwing an exception, but storing 'None' instead.

    Parameters
    ----------
    spectrum
        Spectrum to which to add 'value' in 'target_key'.
    value
        Value to parse into 'target_key'.
    target_key
        Name of the key in which to store the value.

    Returns
    -------
    Spectrum with added key.
    """
    if value is not None:   # one of accepted keys is present
        value = safe_convert_to_float(value)
    spectrum.set(target_key, value)
    return spectrum


def safe_convert_to_float(value: Any) -> Optional[float]:
    """Safely convert value to float. Return 'None' on failure.

    Parameters
    ----------
    value
        Object to convert to float.

    Returns
    -------
    Converted float value or 'None' if conversion is not possible.
    """
    if isinstance(value, list) and len(value) == 1:
        value = value[0]
    try:
        value = float(value)
        rt = value if value >= 0 else None  # discard negative RT values
    except (ValueError, TypeError):
        logger.warning("%s can't be converted to float.", str(value))
        rt = None
    return rt


def _add_retention(spectrum: SpectrumType, target_key: str, accepted_keys: List[str]) -> SpectrumType:
    """Add value from one of accepted keys to target key.

    Parameters
    ----------
    spectrum
        Spectrum from which to read the values.
    target_key
        Key under which to store the value.
    accepted_keys
        List of accepted keys from which a value will be read (in order).

    Returns
    -------
    Spectrum with value from first accepted key stored under target_key.
    """
    common_keys = get_common_keys(spectrum.metadata.keys(), accepted_keys)
    values_for_keys = filter_none([spectrum.get(key) for key in common_keys])
    values = list(map(safe_convert_to_float, values_for_keys))
    value = next(filter_none(values), None)

    spectrum = safe_store_value(spectrum, value, target_key)
    return spectrum


def add_retention_time(spectrum_in: SpectrumType) -> SpectrumType:
    """Add retention time information to the 'retention_time' key as float.
    Negative values and those not convertible to a float result in 'retention_time'
    being 'None'.

    Parameters
    ----------
    spectrum
        Spectrum with retention time information.

    Returns
    -------
    Spectrum with harmonized retention time information.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    target_key = "retention_time"
    spectrum = _add_retention(spectrum, target_key, _retention_time_keys)
    return spectrum


def add_retention_index(spectrum_in: SpectrumType) -> SpectrumType:
    """Add retention index into 'retention_index' key if present.


    Parameters
    ----------
    spectrum
        Spectrum with RI information.
    Returns
    -------
    Spectrum with RI info stored under 'retention_index'.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    target_key = "retention_index"
    spectrum = _add_retention(spectrum, target_key, _retention_index_keys)
    return spectrum



import numpy
from ..Spikes import Spikes
from ..typing import SpectrumType


def remove_peaks_around_precursor_mz(spectrum_in: SpectrumType, mz_tolerance: float = 17) -> SpectrumType:

    """Remove peaks that are within mz_tolerance (in Da) of
       the precursor mz, exlcuding the precursor peak.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    mz_tolerance:
        Tolerance of mz values that are not allowed to lie
        within the precursor mz. Default is 17 Da.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    precursor_mz = spectrum.get("precursor_mz", None)
    assert precursor_mz is not None, "Precursor mz absent."
    assert isinstance(precursor_mz, (float, int)), ("Expected 'precursor_mz' to be a scalar number.",
                                                    "Consider applying 'add_precursor_mz' filter first.")
    assert mz_tolerance >= 0, "mz_tolerance must be a positive scalar."

    mzs, intensities = spectrum.peaks.mz, spectrum.peaks.intensities
    peaks_to_remove = ((numpy.abs(precursor_mz-mzs) <= mz_tolerance) & (mzs != precursor_mz))
    new_mzs, new_intensities = mzs[~peaks_to_remove], intensities[~peaks_to_remove]
    spectrum.peaks = Spikes(mz=new_mzs, intensities=new_intensities)

    return spectrum



from ..typing import SpectrumType


def set_ionmode_na_when_missing(spectrum_in: SpectrumType) -> SpectrumType:
    """Create "ionmode" entry of "n/a" if field does not yet exist in metadata."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    # Set ionmode to "n/a" when ionmode is missing from the metadata
    if spectrum.get("ionmode") is None:
        spectrum.set("ionmode", "n/a")

    return spectrum



"""Helper functions for parsing metadata.
"""
from typing import Any
from typing import Union


def find_by_key(data: Union[list, dict], target: str) -> Any:
    """Helper function to return entries from nested list/dictionary.

    Parameters
    ----------
    data:
        Nested dictionary or list in which entry should be searched.
    target:
        Name of field to search for in data.
    """
    if hasattr(data, "items"):
        for key, value in data.items():
            if key == target:
                yield value
            if isinstance(value, dict):
                yield from find_by_key(value, target)
            elif isinstance(value, list):
                for val in value:
                    yield from find_by_key(val, target)

    elif isinstance(data, list):
        for subdata in data:
            yield from find_by_key(subdata, target)


def parse_mzml_mzxml_metadata(spectrum_dict: dict) -> dict:
    """Parse relevant mzml (or mzxml) metadata entries.

    Parameters
    ----------
    spectrum_dict:
        Spectrum dictionary containing metadata fields. Metadata parsing may easily
        break when field key names vary. The following metadata information is considered
        here:

        - precursor_mz, searched for in:
            -->"precursor"/"precursorMz"--> ... --> "selected ion m/z"/"precursorMz"
        - charge, searched for in:
            --> "charge state"/"polarity"
        - title, searched for in "spectrum title"
        - scan_number, searched for in "num"
        - scan_start_time, searched for in "scan start time"
        - retention_time, searched for in "retentionTime"

    """
    charge = None
    title = None
    scan_number = None
    precursor_mz = None
    scan_time = None
    retention_time = None

    first_search = list(find_by_key(spectrum_dict, "precursor"))
    if not first_search:
        first_search = list(find_by_key(spectrum_dict, "precursorMz"))
    if first_search:
        precursor_mz_search = list(find_by_key(first_search, "selected ion m/z"))
        if not precursor_mz_search:
            precursor_mz_search = list(find_by_key(first_search, "precursorMz"))
        if precursor_mz_search:
            precursor_mz = float(precursor_mz_search[0])

    precursor_charge = list(find_by_key(first_search, "charge state"))
    if precursor_charge:
        charge = int(precursor_charge[0])
    elif "polarity" in spectrum_dict:
        if spectrum_dict["polarity"] == "-":
            charge = -1
        elif spectrum_dict["polarity"] == "+":
            charge = 1

    if "spectrum title" in spectrum_dict:
        title = spectrum_dict["spectrum title"]
    if "num" in spectrum_dict:
        scan_number = spectrum_dict["num"]

    scan_time = list(find_by_key(spectrum_dict, "scan start time"))
    retention_time = list(find_by_key(spectrum_dict, "retentionTime"))

    return {"charge": charge,
            "scan_number": scan_number,
            "title": title,
            "precursor_mz": precursor_mz,
            "scan_start_time": scan_time,
            "retention_time": retention_time}



import logging
from ..importing import load_adducts_dict
from ..typing import SpectrumType
from ..utils import clean_adduct


logger = logging.getLogger("matchms")


def derive_ionmode(spectrum_in: SpectrumType) -> SpectrumType:
    """Derive missing ionmode based on adduct.

    Some input formates (e.g. MGF files) do not always provide a correct ionmode.
    This function reads the adduct from the metadata and uses this to fill in the
    correct ionmode where missing.

    Parameters
    ----------
    spectrum
        Input spectrum.

    Returns
    -------
    Spectrum object with `ionmode` attribute set.
    """

    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    # Load lists of known adducts
    known_adducts = load_adducts_dict()

    adduct = spectrum.get("adduct", None)
    # Harmonize adduct string
    if adduct:
        adduct = clean_adduct(adduct)

    ionmode = spectrum.get("ionmode")
    if ionmode:
        assert ionmode == ionmode.lower(), ("Ionmode field not harmonized.",
                                            "Apply 'make_ionmode_lowercase' filter first.")
    if ionmode in ["positive", "negative"]:
        return spectrum

    # Try completing missing or incorrect ionmodes
    if adduct in known_adducts:
        ionmode = known_adducts[adduct]["ionmode"]
    else:
        ionmode = "n/a"

    spectrum.set("ionmode", ionmode)
    logger.info("Set ionmode to %s.", ionmode)

    return spectrum



from typing import List
from ..typing import SpectrumType


def harmonize_undefined_smiles(spectrum_in: SpectrumType, undefined: str = "",
                               aliases: List[str] = None) -> SpectrumType:
    """Replace all aliases for empty/undefined smiles entries by ``undefined``.

    Parameters
    ----------
    undefined:
        Give desired entry for undefined smiles fields. Default is "".
    aliases:
        Enter list of strings that are expected to represent undefined entries.
        Default is ["", "N/A", "NA", "n/a", "no data"].
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    if aliases is None:
        aliases = [
            "",
            "N/A",
            "NA",
            "n/a",
            "no data"
        ]

    smiles = spectrum.get("smiles")
    if smiles is None:
        # spectrum does not have a "smiles" key in its metadata
        spectrum.set("smiles", undefined)
        return spectrum

    if smiles in aliases:
        # harmonize aliases for undefined values
        spectrum.set("smiles", undefined)

    return spectrum



import logging
from ..typing import SpectrumType
from ..utils import convert_inchi_to_inchikey
from ..utils import is_valid_inchi
from ..utils import is_valid_inchikey


logger = logging.getLogger("matchms")


def derive_inchikey_from_inchi(spectrum_in: SpectrumType) -> SpectrumType:
    """Find missing InchiKey and derive from Inchi where possible."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()
    inchi = spectrum.get("inchi")
    inchikey = spectrum.get("inchikey")

    if is_valid_inchi(inchi) and not is_valid_inchikey(inchikey):
        inchikey = convert_inchi_to_inchikey(inchi)
        if inchikey:
            spectrum.set("inchikey", inchikey)
            logger.info("Added InChIKey %s to metadata (was converted from inchi)", inchikey)
        else:
            logger.warning("Could not convert InChI %s to inchikey.", inchi)

    return spectrum



import re


class SpeciesString:

    def __init__(self, dirty: str):
        self.dirty = dirty
        self.target = None
        self.cleaned = None
        self.guess_target()
        self.clean()

    def __str__(self):
        if self.cleaned == "":
            return ""
        return f"({self.target}): {self.cleaned}"

    def clean(self):

        if self.target is None:
            self.cleaned = ""

        elif self.target == "inchi":
            self.clean_as_inchi()

        elif self.target == "inchikey":
            self.clean_as_inchikey()

        elif self.target == "smiles":
            self.clean_as_smiles()

        return self

    def clean_as_inchi(self):
        """Search for valid inchi and harmonize it."""
        regexp = r"(1S\/|1\/)[0-9, A-Z, a-z,\.]{2,}\/(c|h)[0-9].*$"
        found = re.search(regexp, self.dirty)
        if found is None:
            self.cleaned = ""
        else:
            self.cleaned = "InChI=" + found[0].replace('"', "")

    def clean_as_inchikey(self):
        """Search for valid inchikey and harmonize it."""
        regexp = r"[A-Z]{14}-[A-Z]{10}-[A-Z]"
        found = re.search(regexp, self.dirty)
        if found is None:
            self.cleaned = ""
        else:
            self.cleaned = found[0]

    def clean_as_smiles(self):
        """Search for valid smiles and harmonize it."""
        regexp = r"^([^J][0-9BCOHNSOPIFKcons@+\-\[\]\(\)\\\/%=#$,.~&!|Si|Se|Br|Mg|Na|Cl|Al]{3,})$"
        found = re.search(regexp, self.dirty)
        if found is None:
            self.cleaned = ""
        else:
            self.cleaned = found[0]

    def guess_target(self):

        if self.looks_like_an_inchikey():
            self.target = "inchikey"
        elif self.looks_like_an_inchi():
            self.target = "inchi"
        elif self.looks_like_a_smiles():
            self.target = "smiles"
        else:
            self.target = None

        return self

    def looks_like_an_inchi(self):
        """Search for first piece of InChI."""
        regexp = r"(InChI=1|1)(S\/|\/)[0-9, A-Z, a-z,\.]{2,}\/(c|h)[0-9]"
        return re.search(regexp, self.dirty) is not None

    def looks_like_an_inchikey(self):
        """Return True if string has format of inchikey."""
        regexp = r"[A-Z]{14}-[A-Z]{10}-[A-Z]"
        return re.search(regexp, self.dirty) is not None

    def looks_like_a_smiles(self):
        """Return True if string is made of allowed charcters for smiles."""
        regexp = r"^([^J][0-9BCOHNSOPIFKcons@+\-\[\]\(\)\\\/%=#$,.~&!|Si|Se|Br|Mg|Na|Cl|Al]{3,})$"
        return re.search(regexp, self.dirty) is not None



"""
Functions for exporting mass spectral data
##########################################

Individual :meth:`~matchms.Spectrum`, or lists of :meth:`~matchms.Spectrum`
can be exported to json, mgf, or msp files.
"""
from .save_as_json import save_as_json
from .save_as_mgf import save_as_mgf
from .save_as_msp import save_as_msp


__all__ = [
    "save_as_json",
    "save_as_mgf",
    "save_as_msp",
]



import logging
import numpy
from ..typing import SpectrumType
from ..utils import derive_fingerprint_from_inchi
from ..utils import derive_fingerprint_from_smiles


logger = logging.getLogger("matchms")


def add_fingerprint(spectrum_in: SpectrumType, fingerprint_type: str = "daylight",
                    nbits: int = 2048) -> SpectrumType:
    """Add molecular finterprint to spectrum.

    If smiles or inchi present in metadata, derive a molecular finterprint and
    add it to the spectrum.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    fingerprint_type:
        Determine method for deriving molecular fingerprints. Supported choices
        are "daylight", "morgan1", "morgan2", "morgan3". Default is "daylight".
    nbits:
        Dimension or number of bits of generated fingerprint. Default is 2048.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    # First try to get fingerprint from smiles
    if spectrum.get("smiles", None):
        fingerprint = derive_fingerprint_from_smiles(spectrum.get("smiles"),
                                                     fingerprint_type, nbits)
        if isinstance(fingerprint, numpy.ndarray) and fingerprint.sum() > 0:
            spectrum.set("fingerprint", fingerprint)
            return spectrum

    # Second try to get fingerprint from inchi
    if spectrum.get("inchi", None):
        fingerprint = derive_fingerprint_from_inchi(spectrum.get("inchi"),
                                                    fingerprint_type, nbits)
        if isinstance(fingerprint, numpy.ndarray) and fingerprint.sum() > 0:
            spectrum.set("fingerprint", fingerprint)
            return spectrum

    logger.info("No fingerprint was added (name: %s).", spectrum.get("compound_name"))
    return spectrum



"""Helper functions related to hashing.
"""
import hashlib
import json
from .Spikes import Spikes


def spectrum_hash(peaks: Spikes, hash_length: int = 20,
                  mz_precision: int = 5, intensity_precision: int = 2):
    """Compute hash from mz-intensity pairs of all peaks in spectrum.
    Method is inspired by SPLASH (doi:10.1038/nbt.3689).
    """
    mz_precision_factor = 10 ** mz_precision
    intensity_precision_factor = 10 ** intensity_precision

    def format_mz(mz):
        return int(mz * mz_precision_factor)

    def format_intensity(intensity):
        return int(intensity * intensity_precision_factor)

    peak_list = [(format_mz(peak[0]), format_intensity(peak[1])) for peak in peaks.to_numpy]
    # Sort by increasing m/z and then by decreasing intensity
    peak_list.sort(key=lambda x: (x[0], - x[1]))

    encoded = " ".join(":".join(map(str, x)) for x in peak_list).encode("utf-8")
    return hashlib.sha256(encoded).hexdigest()[:hash_length]


def metadata_hash(metadata: dict, hash_length: int = 20):
    """Compute hash from metadata dictionary.
    """
    encoded = json.dumps(metadata, sort_keys=True).encode()
    return hashlib.sha256(encoded).hexdigest()[:hash_length]



import logging
import numpy
from matchms.typing import SpectrumType
from ..Spikes import Spikes


logger = logging.getLogger("matchms")


def normalize_intensities(spectrum_in: SpectrumType) -> SpectrumType:
    """Normalize intensities of peaks (and losses) to unit height."""

    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    if len(spectrum.peaks) == 0:
        return spectrum

    max_intensity = numpy.max(spectrum.peaks.intensities)

    if max_intensity <= 0:
        logger.warning("Spectrum with all peak intensities <= 0 was set to None.")
        return None

    # Normalize peak intensities
    mz, intensities = spectrum.peaks.mz, spectrum.peaks.intensities
    normalized_intensities = intensities / max_intensity
    spectrum.peaks = Spikes(mz=mz, intensities=normalized_intensities)

    # Normalize loss intensities
    if spectrum.losses is not None and len(spectrum.losses) > 0:
        mz, intensities = spectrum.losses.mz, spectrum.losses.intensities
        normalized_intensities = intensities / max_intensity
        spectrum.losses = Spikes(mz=mz, intensities=normalized_intensities)

    return spectrum



import numpy
from ..typing import SpectrumType


def correct_charge(spectrum_in: SpectrumType) -> SpectrumType:
    """Correct charge values based on given ionmode.

    For some spectrums, the charge value is either undefined or inconsistent with its
    ionmode, which is corrected by this filter.

    Parameters
    ----------
    spectrum_in
        Input spectrum.
    """

    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    ionmode = spectrum.get("ionmode", None)
    if ionmode:
        assert ionmode == ionmode.lower(), ("Ionmode field not harmonized.",
                                            "Apply 'make_ionmode_lowercase' filter first.")

    charge = spectrum.get("charge", None)
    assert not isinstance(charge, str), ("Charge is given as string.",
                                         "Apply 'make_charge_scalar' filter first.")

    if charge is None:
        charge = 0

    if charge == 0 and ionmode == 'positive':
        charge = 1
    elif charge == 0 and ionmode == 'negative':
        charge = -1

    # Correct charge when in conflict with ionmode (trust ionmode more!)
    if numpy.sign(charge) == 1 and ionmode == 'negative':
        charge *= -1
    elif numpy.sign(charge) == -1 and ionmode == 'positive':
        charge *= -1

    spectrum.set("charge", charge)

    return spectrum



import logging
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def make_charge_int(spectrum_in: SpectrumType) -> SpectrumType:
    """Convert charge field to integer (if possible)."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    # Avoid pyteomics ChargeList
    if isinstance(spectrum.get("charge", None), list):
        spectrum.set("charge", int(spectrum.get("charge")[0]))

    # convert string charges to int
    if isinstance(spectrum.get("charge", None), str):
        try:
            charge_int = int(spectrum.get('charge'))
            spectrum.set("charge", charge_int)
        except ValueError:
            logger.warning("Found charge (%s) cannot be converted to integer.", str(spectrum.get('charge')))

    return spectrum



"""Collection of functions for calculating vector-vector similarities."""
import numba
import numpy


@numba.njit
def jaccard_similarity_matrix(references: numpy.ndarray, queries: numpy.ndarray) -> numpy.ndarray:
    """Returns matrix of jaccard indices between all-vs-all vectors of references
    and queries.

    Parameters
    ----------
    references
        Reference vectors as 2D numpy array. Expects that vector_i corresponds to
        references[i, :].
    queries
        Query vectors as 2D numpy array. Expects that vector_i corresponds to
        queries[i, :].

    Returns
    -------
    scores
        Matrix of all-vs-all similarity scores. scores[i, j] will contain the score
        between the vectors references[i, :] and queries[j, :].
    """
    size1 = references.shape[0]
    size2 = queries.shape[0]
    scores = numpy.zeros((size1, size2))
    for i in range(size1):
        for j in range(size2):
            scores[i, j] = jaccard_index(references[i, :], queries[j, :])
    return scores


@numba.njit
def dice_similarity_matrix(references: numpy.ndarray, queries: numpy.ndarray) -> numpy.ndarray:
    """Returns matrix of dice similarity scores between all-vs-all vectors of references
    and queries.

    Parameters
    ----------
    references
        Reference vectors as 2D numpy array. Expects that vector_i corresponds to
        references[i, :].
    queries
        Query vectors as 2D numpy array. Expects that vector_i corresponds to
        queries[i, :].

    Returns
    -------
    scores
        Matrix of all-vs-all similarity scores. scores[i, j] will contain the score
        between the vectors references[i, :] and queries[j, :].
    """
    size1 = references.shape[0]
    size2 = queries.shape[0]
    scores = numpy.zeros((size1, size2))
    for i in range(size1):
        for j in range(size2):
            scores[i, j] = dice_similarity(references[i, :], queries[j, :])
    return scores


@numba.njit
def cosine_similarity_matrix(references: numpy.ndarray, queries: numpy.ndarray) -> numpy.ndarray:
    """Returns matrix of cosine similarity scores between all-vs-all vectors of
    references and queries.

    Parameters
    ----------
    references
        Reference vectors as 2D numpy array. Expects that vector_i corresponds to
        references[i, :].
    queries
        Query vectors as 2D numpy array. Expects that vector_i corresponds to
        queries[i, :].

    Returns
    -------
    scores
        Matrix of all-vs-all similarity scores. scores[i, j] will contain the score
        between the vectors references[i, :] and queries[j, :].
    """
    size1 = references.shape[0]
    size2 = queries.shape[0]
    scores = numpy.zeros((size1, size2))
    for i in range(size1):
        for j in range(size2):
            scores[i, j] = cosine_similarity(references[i, :], queries[j, :])
    return scores


@numba.njit
def jaccard_index(u: numpy.ndarray, v: numpy.ndarray) -> numpy.float64:
    r"""Computes the Jaccard-index (or Jaccard similarity coefficient) of two boolean
    1-D arrays.
    The Jaccard index between 1-D boolean arrays `u` and `v`,
    is defined as

    .. math::

       J(u,v) = \\frac{u \cap v}
                {u \cup v}

    Parameters
    ----------
    u :
        Input array. Expects boolean vector.
    v :
        Input array. Expects boolean vector.

    Returns
    -------
    jaccard_similarity
        The Jaccard similarity coefficient between vectors `u` and `v`.
    """
    u_or_v = numpy.bitwise_or(u != 0, v != 0)
    u_and_v = numpy.bitwise_and(u != 0, v != 0)
    jaccard_score = 0
    if u_or_v.sum() != 0:
        jaccard_score = numpy.float64(u_and_v.sum()) / numpy.float64(u_or_v.sum())
    return jaccard_score


@numba.njit
def dice_similarity(u: numpy.ndarray, v: numpy.ndarray) -> numpy.float64:
    r"""Computes the Dice similarity coefficient (DSC) between two boolean 1-D arrays.

    The Dice similarity coefficient between `u` and `v`, is

    .. math::

         DSC(u,v) = \\frac{2|u \cap v|}
                    {|u| + |v|}

    Parameters
    ----------
    u
        Input array. Expects boolean vector.
    v
        Input array. Expects boolean vector.

    Returns
    -------
    dice_similarity
        The Dice similarity coefficient between 1-D arrays `u` and `v`.
    """
    u_and_v = numpy.bitwise_and(u != 0, v != 0)
    u_abs_and_v_abs = numpy.abs(u).sum() + numpy.abs(v).sum()
    dice_score = 0
    if u_abs_and_v_abs != 0:
        dice_score = 2.0 * numpy.float64(u_and_v.sum()) / numpy.float64(u_abs_and_v_abs)
    return dice_score


@numba.njit
def cosine_similarity(u: numpy.ndarray, v: numpy.ndarray) -> numpy.float64:
    """Calculate cosine similarity score.

    Parameters
    ----------
    u
        Input vector.
    v
        Input vector.

    Returns
    -------
    cosine_similarity
        The Cosine similarity score between vectors `u` and `v`.
    """
    assert u.shape[0] == v.shape[0], "Input vector must have same shape."
    uv = 0
    uu = 0
    vv = 0
    for i in range(u.shape[0]):
        uv += u[i] * v[i]
        uu += u[i] * u[i]
        vv += v[i] * v[i]
    cosine_score = 0
    if uu != 0 and vv != 0:
        cosine_score = uv / numpy.sqrt(uu * vv)
    return numpy.float64(cosine_score)



import numpy
from ..Spikes import Spikes
from ..typing import SpectrumType


def remove_peaks_outside_top_k(spectrum_in: SpectrumType, k: int = 6,
                               mz_window: float = 50) -> SpectrumType:

    """Remove all peaks which are not within *mz_window* of at least one
       of the *k* highest intensity peaks of the spectrum.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    k:
        The number of most intense peaks to compare to. Default is 6.
    mz_window:
        Window of mz values (in Da) that are allowed to lie within
        the top k peaks. Default is 50 Da.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    assert k >= 1, "k must be a positive nonzero integer."
    assert mz_window >= 0, "mz_window must be a positive scalar."
    mzs, intensities = spectrum.peaks.mz, spectrum.peaks.intensities
    top_k = intensities.argsort()[::-1][0:k]
    k_ordered_mzs = mzs[top_k]
    indices = [i for i in range(len(mzs)) if i not in top_k]
    keep_idx = top_k.tolist()
    for i in indices:

        compare = abs(mzs[i]-k_ordered_mzs) <= mz_window
        if numpy.any(compare):
            keep_idx.append(i)

    keep_idx.sort()
    new_mzs, new_intensities = mzs[keep_idx], intensities[keep_idx]
    spectrum.peaks = Spikes(mz=new_mzs, intensities=new_intensities)

    return spectrum



import logging
from typing import Tuple
import numpy
from matchms.filtering.add_precursor_mz import _convert_precursor_mz
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity
from .spectrum_similarity_functions import collect_peak_pairs
from .spectrum_similarity_functions import score_best_matches


logger = logging.getLogger("matchms")


class ModifiedCosine(BaseSimilarity):
    """Calculate 'modified cosine score' between mass spectra.

    The modified cosine score aims at quantifying the similarity between two
    mass spectra. The score is calculated by finding best possible matches between
    peaks of two spectra. Two peaks are considered a potential match if their
    m/z ratios lie within the given 'tolerance', or if their m/z ratios
    lie within the tolerance once a mass-shift is applied. The mass shift is
    simply the difference in precursor-m/z between the two spectra.
    See Watrous et al. [PNAS, 2012, https://www.pnas.org/content/109/26/E1743]
    for further details.

    For example

    .. testcode::

        import numpy as np
        from matchms import Spectrum
        from matchms.similarity import ModifiedCosine

        spectrum_1 = Spectrum(mz=np.array([100, 150, 200.]),
                              intensities=np.array([0.7, 0.2, 0.1]),
                              metadata={"precursor_mz": 100.0})
        spectrum_2 = Spectrum(mz=np.array([104.9, 140, 190.]),
                              intensities=np.array([0.4, 0.2, 0.1]),
                              metadata={"precursor_mz": 105.0})

        # Use factory to construct a similarity function
        modified_cosine = ModifiedCosine(tolerance=0.2)

        score = modified_cosine.pair(spectrum_1, spectrum_2)

        print(f"Modified cosine score is {score['score']:.2f} with {score['matches']} matched peaks")

    Should output

    .. testoutput::

        Modified cosine score is 0.83 with 1 matched peaks

    """
    # Set key characteristics as class attributes
    is_commutative = True
    # Set output data type, e.g. ("score", "float") or [("score", "float"), ("matches", "int")]
    score_datatype = [("score", numpy.float64), ("matches", "int")]

    def __init__(self, tolerance: float = 0.1, mz_power: float = 0.0,
                 intensity_power: float = 1.0):
        """
        Parameters
        ----------
        tolerance:
            Peaks will be considered a match when <= tolerance apart. Default is 0.1.
        mz_power:
            The power to raise mz to in the cosine function. The default is 0, in which
            case the peak intensity products will not depend on the m/z ratios.
        intensity_power:
            The power to raise intensity to in the cosine function. The default is 1.
        """
        self.tolerance = tolerance
        self.mz_power = mz_power
        self.intensity_power = intensity_power

    def pair(self, reference: SpectrumType, query: SpectrumType) -> Tuple[float, int]:
        """Calculate modified cosine score between two spectra.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.

        Returns
        -------

        Tuple with cosine score and number of matched peaks.
        """
        def get_valid_precursor_mz(spectrum):
            """Extract valid precursor_mz from spectrum if possible. If not raise exception."""
            message_precursor_missing = \
                "Precursor_mz missing. Apply 'add_precursor_mz' filter first."
            message_precursor_no_number = \
                "Precursor_mz must be of type int or float. Apply 'add_precursor_mz' filter first."
            message_precursor_below_0 = "Expect precursor to be positive number." \
                                        "Apply 'require_precursor_mz' first"

            precursor_mz = spectrum.get("precursor_mz", None)
            if not isinstance(precursor_mz, (int, float)):
                logger.warning(message_precursor_no_number)
            precursor_mz = _convert_precursor_mz(precursor_mz)
            assert precursor_mz is not None, message_precursor_missing
            assert precursor_mz > 0, message_precursor_below_0
            return precursor_mz

        def get_matching_pairs():
            """Find all pairs of peaks that match within the given tolerance."""
            zero_pairs = collect_peak_pairs(spec1, spec2, self.tolerance, shift=0.0,
                                            mz_power=self.mz_power,
                                            intensity_power=self.intensity_power)
            precursor_mz_ref = get_valid_precursor_mz(reference)
            precursor_mz_query = get_valid_precursor_mz(query)

            mass_shift = precursor_mz_ref - precursor_mz_query
            nonzero_pairs = collect_peak_pairs(spec1, spec2, self.tolerance, shift=mass_shift,
                                               mz_power=self.mz_power,
                                               intensity_power=self.intensity_power)

            if zero_pairs is None:
                zero_pairs = numpy.zeros((0, 3))
            if nonzero_pairs is None:
                nonzero_pairs = numpy.zeros((0, 3))
            matching_pairs = numpy.concatenate((zero_pairs, nonzero_pairs), axis=0)
            if matching_pairs.shape[0] > 0:
                matching_pairs = matching_pairs[numpy.argsort(matching_pairs[:, 2])[::-1], :]
            return matching_pairs

        spec1 = reference.peaks.to_numpy
        spec2 = query.peaks.to_numpy
        matching_pairs = get_matching_pairs()
        if matching_pairs.shape[0] == 0:
            return numpy.asarray((float(0), 0), dtype=self.score_datatype)
        score = score_best_matches(matching_pairs, spec1, spec2,
                                   self.mz_power, self.intensity_power)
        return numpy.asarray(score, dtype=self.score_datatype)



from typing import Generator
import numpy
from pyteomics import mzml
from matchms.importing.parsing_utils import parse_mzml_mzxml_metadata
from matchms.Spectrum import Spectrum


def load_from_mzml(filename: str, ms_level: int = 2) -> Generator[Spectrum, None, None]:
    """Load spectrum(s) from mzml file.

    This function will create ~matchms.Spectrum for every spectrum of desired
    ms_level found in a given MzML file. For more extensive parsing options consider
    using pyteomics or pymzml packages.

    Example:

    .. code-block:: python

        from matchms.importing import load_from_mzml

        file_mzml = "testdata.mzml"
        spectrums = list(load_from_mzml(file_mzml))

    Parameters
    ----------
    filename:
        Filename for mzml file to import.
    ms_level:
        Specify which ms level to import. Default is 2.
    """
    for pyteomics_spectrum in mzml.read(filename, dtype=dict):
        if "ms level" in pyteomics_spectrum and pyteomics_spectrum["ms level"] == ms_level:
            metadata = parse_mzml_mzxml_metadata(pyteomics_spectrum)
            mz = numpy.asarray(pyteomics_spectrum["m/z array"], dtype="float")
            intensities = numpy.asarray(pyteomics_spectrum["intensity array"], dtype="float")

            if mz.shape[0] > 0:
                # Sort by mz (if not sorted already)
                if not numpy.all(mz[:-1] <= mz[1:]):
                    idx_sorted = numpy.argsort(mz)
                    mz = mz[idx_sorted]
                    intensities = intensities[idx_sorted]

                yield Spectrum(mz=mz, intensities=intensities, metadata=metadata)



from . import exporting
from . import filtering
from . import importing
from . import networking
from . import plotting
from . import similarity
from .__version__ import __version__
from .calculate_scores import calculate_scores
from .logging_functions import _init_logger
from .logging_functions import set_matchms_logger_level
from .Scores import Scores
from .Spectrum import Spectrum
from .Spikes import Spikes


_init_logger()


__author__ = "Netherlands eScience Center"
__email__ = 'generalization@esciencecenter.nl'
__all__ = [
    "__version__",
    "calculate_scores",
    "exporting",
    "filtering",
    "importing",
    "networking",
    "plotting",
    "Scores",
    "set_matchms_logger_level",
    "similarity",
    "Spectrum",
    "Spikes"
]



"""
Functions for creating and analysing spectral networks
######################################################

"""
from .SimilarityNetwork import SimilarityNetwork


__all__ = [
    "SimilarityNetwork",
]



from typing import List
import pyteomics.mgf as py_mgf
from ..Spectrum import Spectrum


def save_as_mgf(spectrums: List[Spectrum], filename: str):
    """Save spectrum(s) as mgf file.

    :py:attr:`~matchms.Spectrum.losses` of spectrum will not be saved.

    Example:

    .. code-block:: python

        import numpy
        from matchms import Spectrum
        from matchms.exporting import save_as_mgf

        # Create dummy spectrum
        spectrum = Spectrum(mz=numpy.array([100, 200, 300], dtype="float"),
                            intensities=numpy.array([10, 10, 500], dtype="float"),
                            metadata={"charge": -1,
                                      "inchi": '"InChI=1S/C6H12"',
                                      "precursor_mz": 222.2})

        # Write spectrum to test file
        save_as_mgf(spectrum, "test.mgf")

    Parameters
    ----------
    spectrums:
        Expected input are match.Spectrum.Spectrum() objects.
    filename:
        Provide filename to save spectrum(s).
    """
    if not isinstance(spectrums, list):
        # Assume that input was single Spectrum
        spectrums = [spectrums]

    # Convert matchms.Spectrum() into dictionaries for pyteomics
    for spectrum in spectrums:
        spectrum_dict = {"m/z array": spectrum.peaks.mz,
                         "intensity array": spectrum.peaks.intensities,
                         "params": spectrum.metadata}
        # Append spectrum to file
        py_mgf.write([spectrum_dict], filename)



import logging
import numpy
from ..Spikes import Spikes
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def add_losses(spectrum_in: SpectrumType, loss_mz_from=0.0, loss_mz_to=1000.0) -> SpectrumType:
    """Derive losses based on precursor mass.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    loss_mz_from:
        Minimum allowed m/z value for losses. Default is 0.0.
    loss_mz_to:
        Maximum allowed m/z value for losses. Default is 1000.0.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    precursor_mz = spectrum.get("precursor_mz", None)
    if precursor_mz:
        assert isinstance(precursor_mz, (float, int)), ("Expected 'precursor_mz' to be a scalar number.",
                                                        "Consider applying 'add_precursor_mz' filter first.")
        peaks_mz, peaks_intensities = spectrum.peaks.mz, spectrum.peaks.intensities
        losses_mz = (precursor_mz - peaks_mz)[::-1]
        losses_intensities = peaks_intensities[::-1]
        # Add losses which are within given boundaries
        mask = numpy.where((losses_mz >= loss_mz_from)
                           & (losses_mz <= loss_mz_to))
        spectrum.losses = Spikes(mz=losses_mz[mask],
                                 intensities=losses_intensities[mask])
    else:
        logger.warning("No precursor_mz found. Consider applying 'add_precursor_mz' filter first.")

    return spectrum



import logging
from ..typing import SpectrumType
from ..utils import convert_inchi_to_smiles
from ..utils import is_valid_inchi
from ..utils import is_valid_smiles


logger = logging.getLogger("matchms")


def derive_smiles_from_inchi(spectrum_in: SpectrumType) -> SpectrumType:
    """Find missing smiles and derive from Inchi where possible."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()
    inchi = spectrum.get("inchi")
    smiles = spectrum.get("smiles")

    if not is_valid_smiles(smiles) and is_valid_inchi(inchi):
        smiles = convert_inchi_to_smiles(inchi)
        if smiles:
            smiles = smiles.rstrip()
            spectrum.set("smiles", smiles)
            logger.info("Added smiles %s to metadata (was converted from InChI)", smiles)
        else:
            logger.warning("Could not convert InChI %s to smiles.", inchi)

    return spectrum



import numpy


class Spikes:
    """
    Stores arrays of intensities and M/z values, with some checks on their internal consistency.

    For example

    .. testcode::

        import numpy as np
        from matchms import Spikes

        mz = np.array([10, 20, 30], dtype="float")
        intensities = np.array([100, 20, 300], dtype="float")

        peaks = Spikes(mz=mz, intensities=intensities)
        print(peaks[2])

    Should output

    .. testoutput::

       [ 30. 300.]

    Attributes
    ----------
    mz:
        Numpy array of m/z values.
    intensities:
        Numpy array of peak intensity values.

    """
    def __init__(self, mz=None, intensities=None):
        assert isinstance(mz, numpy.ndarray), "Input argument 'mz' should be a numpy.array."
        assert isinstance(intensities, numpy.ndarray), "Input argument 'intensities' should be a numpy.array."
        assert mz.shape == intensities.shape, "Input arguments 'mz' and 'intensities' should be the same shape."
        assert mz.dtype == "float", "Input argument 'mz' should be an array of type float."
        assert intensities.dtype == "float", "Input argument 'intensities' should be an array of type float."

        self._mz = mz
        self._intensities = intensities

        assert self._is_sorted(), "mz values are out of order."

    def __eq__(self, other):
        return \
            self.mz.shape == other.mz.shape and \
            numpy.allclose(self.mz, other.mz) and \
            self.intensities.shape == other.intensities.shape and \
            numpy.allclose(self.intensities, other.intensities)

    def __len__(self):
        return self._mz.size

    def __getitem__(self, item):
        return numpy.asarray([self.mz[item], self.intensities[item]])

    def _is_sorted(self):
        return numpy.all(self.mz[:-1] <= self.mz[1:])

    def clone(self):
        return Spikes(self.mz, self.intensities)

    @property
    def mz(self):
        """getter method for mz private variable"""
        return self._mz.copy()

    @property
    def intensities(self):
        """getter method for intensities private variable"""
        return self._intensities.copy()

    @property
    def to_numpy(self):
        """getter method to return stacked numpy array of both peak mz and
        intensities"""
        return numpy.vstack((self.mz, self.intensities)).T



"""
Functions for plotting one or multiple spectra
##############################################

Matchms provides (limited) plotting functionality to display one or multiple
spectra from :class:`~matchms.Spectrum` objects.

Currently this includes 3 different plot types:

* `plot_spectrum(spectrum)` or `spectrum.plot()`
  This will create a plot of an individual spectrum.
* `plot_spectra_mirror(spectrum1, spectrum2)` or `spectrum1.plot_against(spectrum2)`
  This will create a mirro plot comparing two spectra.
* `plot_spectra_array([spec1, spec2, ...])`
  This will create a plot with an array of all spectra in the given list.

Example of how to visually compare two spectra:

.. code-block:: python

    import os
    from matplotlib import pyplot as plt
    import matchms.filtering as msfilters
    from matchms.importing import load_from_msp

    module_root = os.getcwd()
    spectrums_file = os.path.join(module_root, "matchms", "tests", "MoNA-export-GC-MS-first10.msp")
    spectrums = list(load_from_msp(spectrums_file))
    spectrums = [msfilters.default_filters(s) for s in spectrums]

    spectrums[1].plot()
    # plt.savefig("spectrum-plot-example_1.png", dpi=300)  # If you want to save a plot

.. figure:: ../_static/spectrum-plot-example.png
   :width: 700
   :alt: matchms spectrum plot

   Plot of individual spectrum.

Another example is to compare two spectra visually using a mirror plot:

.. code-block:: python

    spectrums[2].plot_against(spectrums[3])
    plt.xlim(0, 200)

.. figure:: ../_static/spectrum-mirror-plot-example.png
   :width: 700
   :alt: matchms spectrum mirror plot

   Compare two spectra visually using a mirror plot.

Finally, it is also possible to plot many spectra at once using `plot_spectra_array()`:

.. code-block:: python

    from matchms.plotting import plot_spectra_array
    plot_spectra_array(spectrums[:4])

.. figure:: ../_static/spectra-array-plot-example.png
   :width: 700
   :alt: matchms spectra array plot

   Compare many spectra visually using an array plot.

"""
from .spectrum_plots import plot_spectra_array
from .spectrum_plots import plot_spectra_mirror
from .spectrum_plots import plot_spectrum


__all__ = [
    "plot_spectrum",
    "plot_spectra_mirror",
    "plot_spectra_array",
]



from typing import Tuple
import numpy
from scipy.optimize import linear_sum_assignment
from matchms.similarity.spectrum_similarity_functions import collect_peak_pairs
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity


class CosineHungarian(BaseSimilarity):
    """Calculate 'cosine similarity score' between two spectra (using Hungarian algorithm).

    The cosine score aims at quantifying the similarity between two mass spectra.
    The score is calculated by finding best possible matches between peaks
    of two spectra. Two peaks are considered a potential match if their
    m/z ratios lie within the given 'tolerance'.
    The underlying peak assignment problem is here solved using the Hungarian algorithm.
    This can perform notably slower than the 'greedy' implementation in
    :class:`~matchms.similarity.CosineGreedy`, but does represent a mathematically proper
    solution to the problem.
    """
    # Set key characteristics as class attributes
    is_commutative = True
    # Set output data type, e.g. ("score", "float") or [("score", "float"), ("matches", "int")]
    score_datatype = [("score", numpy.float64), ("matches", "int")]

    def __init__(self, tolerance: float = 0.1, mz_power: float = 0.0,
                 intensity_power: float = 1.0):
        """
        Parameters
        ----------
        tolerance:
            Peaks will be considered a match when <= tolerance apart. Default is 0.1.
        mz_power:
            The power to raise m/z to in the cosine function. The default is 0, in which
            case the peak intensity products will not depend on the m/z ratios.
        intensity_power:
            The power to raise intensity to in the cosine function. The default is 1.
        """
        self.tolerance = tolerance
        self.mz_power = mz_power
        self.intensity_power = intensity_power

    def pair(self, reference: SpectrumType, query: SpectrumType) -> Tuple[float, int]:
        """Calculate cosine score between two spectra.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.

        Returns:
        --------

        Tuple with cosine score and number of matched peaks.
        """
        def get_matching_pairs():
            """Get pairs of peaks that match within the given tolerance."""
            matching_pairs = collect_peak_pairs(spec1, spec2, self.tolerance, shift=0.0,
                                                mz_power=self.mz_power,
                                                intensity_power=self.intensity_power)
            if matching_pairs is None:
                return None
            matching_pairs = matching_pairs[numpy.argsort(matching_pairs[:, 2])[::-1], :]
            return matching_pairs

        def get_matching_pairs_matrix():
            """Create matrix of multiplied intensities of all matching pairs
            between spectrum1 and spectrum2.
            Returns
            paired_peaks1:
                list of paired peaks of spectrum1
            paired_peaks2:
                list of paired peaks of spectrum2
            matching_pairs_matrix:
                Array of multiplied intensities between all matching peaks.
            """
            if matching_pairs is None:
                return None, None, None
            paired_peaks1 = list(set(matching_pairs[:, 0]))
            paired_peaks2 = list(set(matching_pairs[:, 1]))
            matrix_size = (len(paired_peaks1), len(paired_peaks2))
            matching_pairs_matrix = numpy.ones(matrix_size)
            for i in range(matching_pairs.shape[0]):
                matching_pairs_matrix[paired_peaks1.index(matching_pairs[i, 0]),
                                      paired_peaks2.index(matching_pairs[i, 1])] = 1 - matching_pairs[i, 2]
            return paired_peaks1, paired_peaks2, matching_pairs_matrix

        def solve_hungarian():
            """Use hungarian algorithm to solve the linear sum assignment problem."""
            row_ind, col_ind = linear_sum_assignment(matching_pairs_matrix)
            score = len(row_ind) - matching_pairs_matrix[row_ind, col_ind].sum()
            used_matches = [(paired_peaks1[x], paired_peaks2[y]) for (x, y) in zip(row_ind, col_ind)]
            return score, used_matches

        def calc_score():
            """Calculate cosine similarity score."""
            if matching_pairs_matrix is None:
                return numpy.asarray((0.0, 0), dtype=self.score_datatype)
            score, used_matches = solve_hungarian()
            # Normalize score:
            spec1_power = numpy.power(spec1[:, 0], self.mz_power) \
                * numpy.power(spec1[:, 1], self.intensity_power)
            spec2_power = numpy.power(spec2[:, 0], self.mz_power) \
                * numpy.power(spec2[:, 1], self.intensity_power)
            score = score/(numpy.sqrt(numpy.sum(spec1_power**2)) * numpy.sqrt(numpy.sum(spec2_power**2)))
            return numpy.asarray((score, len(used_matches)), dtype=self.score_datatype)

        spec1 = reference.peaks.to_numpy
        spec2 = query.peaks.to_numpy
        matching_pairs = get_matching_pairs()
        paired_peaks1, paired_peaks2, matching_pairs_matrix = get_matching_pairs_matrix()
        return calc_score()



"""
Functions for processing mass spectra
#####################################

Provided functions will usually only perform a single action to a spectrum.
This can be changes or corrections of metadata, or peak filtering.
More complicated processing pipelines can be build by stacking several of
the provided filters.

Example of how to use a single filter function:

.. testcode::

    import numpy as np
    from matchms import Spectrum
    from matchms.filtering import normalize_intensities

    spectrum = Spectrum(mz=np.array([100, 120, 150, 200.]),
                        intensities=np.array([200.0, 300.0, 50.0, 1.0]),
                        metadata={'id': 'spectrum1'})
    spectrum_filtered = normalize_intensities(spectrum)

    max_intensity = spectrum_filtered.peaks.intensities.max()
    print(f"Maximum intensity is {max_intensity:.2f}")

Should output

.. testoutput::

    Maximum intensity is 1.00

.. figure:: ../_static/filtering_sketch.png
   :width: 700
   :alt: matchms filtering sketch

   Sketch of matchms spectrum processing.

"""
from .add_compound_name import add_compound_name
from .add_fingerprint import add_fingerprint
from .add_losses import add_losses
from .add_parent_mass import add_parent_mass
from .add_precursor_mz import add_precursor_mz
from .add_retention import add_retention_index
from .add_retention import add_retention_time
from .clean_compound_name import clean_compound_name
from .correct_charge import correct_charge
from .default_filters import default_filters
from .derive_adduct_from_name import derive_adduct_from_name
from .derive_formula_from_name import derive_formula_from_name
from .derive_inchi_from_smiles import derive_inchi_from_smiles
from .derive_inchikey_from_inchi import derive_inchikey_from_inchi
from .derive_ionmode import derive_ionmode
from .derive_smiles_from_inchi import derive_smiles_from_inchi
from .harmonize_undefined_inchi import harmonize_undefined_inchi
from .harmonize_undefined_inchikey import harmonize_undefined_inchikey
from .harmonize_undefined_smiles import harmonize_undefined_smiles
from .make_charge_int import make_charge_int
from .make_charge_scalar import make_charge_scalar
from .make_ionmode_lowercase import make_ionmode_lowercase
from .normalize_intensities import normalize_intensities
from .reduce_to_number_of_peaks import reduce_to_number_of_peaks
from .remove_peaks_around_precursor_mz import remove_peaks_around_precursor_mz
from .remove_peaks_outside_top_k import remove_peaks_outside_top_k
from .repair_inchi_inchikey_smiles import repair_inchi_inchikey_smiles
from .require_minimum_number_of_peaks import require_minimum_number_of_peaks
from .require_minimum_of_high_peaks import require_minimum_of_high_peaks
from .require_precursor_below_mz import require_precursor_below_mz
from .require_precursor_mz import require_precursor_mz
from .select_by_intensity import select_by_intensity
from .select_by_mz import select_by_mz
from .select_by_relative_intensity import select_by_relative_intensity
from .set_ionmode_na_when_missing import set_ionmode_na_when_missing
from .SpeciesString import SpeciesString


__all__ = [
    "add_compound_name",
    "add_fingerprint",
    "add_losses",
    "add_parent_mass",
    "add_precursor_mz",
    "add_retention_index",
    "add_retention_time",
    "clean_compound_name",
    "correct_charge",
    "default_filters",
    "derive_adduct_from_name",
    "derive_formula_from_name",
    "derive_inchi_from_smiles",
    "derive_inchikey_from_inchi",
    "derive_ionmode",
    "derive_smiles_from_inchi",
    "harmonize_undefined_inchi",
    "harmonize_undefined_inchikey",
    "harmonize_undefined_smiles",
    "make_charge_int",
    "make_charge_scalar",
    "make_ionmode_lowercase",
    "normalize_intensities",
    "reduce_to_number_of_peaks",
    "remove_peaks_around_precursor_mz",
    "remove_peaks_outside_top_k",
    "repair_inchi_inchikey_smiles",
    "require_minimum_number_of_peaks",
    "require_minimum_of_high_peaks",
    "require_precursor_below_mz",
    "require_precursor_mz",
    "select_by_intensity",
    "select_by_mz",
    "select_by_relative_intensity",
    "set_ionmode_na_when_missing",
    "SpeciesString"
]



import logging
from ..typing import SpectrumType
from .select_by_relative_intensity import select_by_relative_intensity


logger = logging.getLogger("matchms")


def require_minimum_of_high_peaks(spectrum_in: SpectrumType, no_peaks: int = 5,
                                  intensity_percent: float = 2.0) -> SpectrumType:

    """Returns None if the number of peaks with relative intensity
       above or equal to intensity_percent is less than no_peaks.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    no_peaks:
        Minimum number of peaks allowed to have relative intensity
        above intensity_percent. Less peaks will return none.
        Default is 5.
    intensity_percent:
        Minimum relative intensity (as a percentage between 0-100) for
        peaks that are searched. Default is 2
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    assert no_peaks >= 1, "no_peaks must be a positive nonzero integer."
    assert 0 <= intensity_percent <= 100, "intensity_percent must be a scalar between 0-100."
    intensities_above_p = select_by_relative_intensity(spectrum, intensity_from=intensity_percent/100, intensity_to=1.0)
    if len(intensities_above_p.peaks) < no_peaks:
        logger.info("Spectrum with %s (<%s) peaks was set to None.",
                    str(len(intensities_above_p.peaks)), str(no_peaks))
        return None

    return spectrum



"""
Functions for importing mass spectral data
##########################################

Matchms provides a import functions for several commonly used data types, such
as *.mzML*, *.mzXML*, *.mgf*, or *.msp*. It is also possible to load data from
*.json* files (tested for json files from GNPS or json files made with matchms).
Another option is to load spectra based on a unique identifier (USI)
(:meth:`~matchms.importing.load_from_usi`).

For more extensive import options we recommend building custom importers using `pyteomics
<https://github.com/levitsky/pyteomics>`_ or `pymzml <https://github.com/pymzml/pymzML>`_.

To process spectrum metadata, matchms can also make use of known adduct information
which is imported via :mod:`~matchms.importing.load_adducts`.
"""
from .load_adducts import load_adducts_dict
from .load_adducts import load_known_adduct_conversions
from .load_from_json import load_from_json
from .load_from_mgf import load_from_mgf
from .load_from_msp import load_from_msp
from .load_from_mzml import load_from_mzml
from .load_from_mzxml import load_from_mzxml
from .load_from_usi import load_from_usi


__all__ = [
    "load_from_json",
    "load_from_mgf",
    "load_from_msp",
    "load_from_mzml",
    "load_from_mzxml",
    "load_from_usi",
    "load_adducts_dict",
    "load_known_adduct_conversions",
]



from ..typing import SpectrumType


def make_ionmode_lowercase(spectrum_in: SpectrumType) -> SpectrumType:
    """Convert entry in field "ionmode" to lower case."""
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    # if the ionmode key exists in the metadata, lowercase its value
    if spectrum.get("ionmode") is not None:
        spectrum.set("ionmode", spectrum.get("ionmode").lower())

    return spectrum



from typing import List
from typing import Union
import numpy
from matchms.typing import SpectrumType
from .BaseSimilarity import BaseSimilarity
from .vector_similarity_functions import cosine_similarity
from .vector_similarity_functions import cosine_similarity_matrix
from .vector_similarity_functions import dice_similarity
from .vector_similarity_functions import dice_similarity_matrix
from .vector_similarity_functions import jaccard_index
from .vector_similarity_functions import jaccard_similarity_matrix


class FingerprintSimilarity(BaseSimilarity):
    """Calculate similarity between molecules based on their fingerprints.

    For this similarity measure to work, fingerprints are expected to be derived
    by running :meth:`~matchms.filtering.add_fingerprint`.

    Code example:

    .. testcode::

        import numpy as np
        from matchms import calculate_scores
        from matchms import Spectrum
        from matchms.filtering import add_fingerprint
        from matchms.similarity import FingerprintSimilarity

        spectrum_1 = Spectrum(mz=np.array([], dtype="float"),
                              intensities=np.array([], dtype="float"),
                              metadata={"smiles": "CCC(C)C(C(=O)O)NC(=O)CCl"})

        spectrum_2 = Spectrum(mz=np.array([], dtype="float"),
                              intensities=np.array([], dtype="float"),
                              metadata={"smiles": "CC(C)C(C(=O)O)NC(=O)CCl"})

        spectrum_3 = Spectrum(mz=np.array([], dtype="float"),
                              intensities=np.array([], dtype="float"),
                              metadata={"smiles": "C(C(=O)O)(NC(=O)O)S"})

        spectrums = [spectrum_1, spectrum_2, spectrum_3]
        # Add fingerprints
        spectrums = [add_fingerprint(x, nbits=256) for x in spectrums]

        # Specify type and calculate similarities
        similarity_measure = FingerprintSimilarity("jaccard")
        scores = calculate_scores(spectrums, spectrums, similarity_measure)
        print(np.round(scores.scores, 3))

    Should output

    .. testoutput::

        [[1.    0.878 0.415]
         [0.878 1.    0.444]
         [0.415 0.444 1.   ]]

    """
    # Set key characteristics as class attributes
    is_commutative = True
    # Set output data type, e.g.  "float" or [("score", "float"), ("matches", "int")]
    score_datatype = numpy.float64

    def __init__(self, similarity_measure: str = "jaccard",
                 set_empty_scores: Union[float, int, str] = "nan"):
        """

        Parameters
        ----------
        similarity_measure:
            Chose similarity measure form "cosine", "dice", "jaccard".
            The default is "jaccard".
        set_empty_scores:
            Define what should be given instead of a similarity score in cases
            where fingprints are missing. The default is "nan", which will return
            numpy.nan's in such cases.
        """
        self.set_empty_scores = set_empty_scores
        assert similarity_measure in ["cosine", "dice", "jaccard"], "Unknown similarity measure."
        self.similarity_measure = similarity_measure

    def pair(self, reference: SpectrumType, query: SpectrumType) -> float:
        """Calculate fingerprint based similarity score between two spectra.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.
        """
        fingerprint_ref = reference.get("fingerprint")
        fingerprint_query = query.get("fingerprint")
        if self.similarity_measure == "jaccard":
            return jaccard_index(fingerprint_ref, fingerprint_query)

        if self.similarity_measure == "dice":
            return dice_similarity(fingerprint_ref, fingerprint_query)

        if self.similarity_measure == "cosine":
            score = cosine_similarity(fingerprint_ref, fingerprint_query)
            return numpy.asarray(score, dtype=self.score_datatype)

        raise NotImplementedError

    def matrix(self, references: List[SpectrumType], queries: List[SpectrumType],
               is_symmetric: bool = False) -> numpy.array:
        """Calculate matrix of fingerprint based similarity scores.

        Parameters
        ----------
        references:
            List of reference spectrums.
        queries:
            List of query spectrums.
        """
        def get_fingerprints(spectrums):
            for index, spectrum in enumerate(spectrums):
                yield index, spectrum.get("fingerprint")

        def collect_fingerprints(spectrums):
            """Collect fingerprints and indices of spectrum with finterprints."""
            idx_fingerprints = []
            fingerprints = []
            for index, fp in get_fingerprints(spectrums):
                if fp is not None:
                    idx_fingerprints.append(index)
                    fingerprints.append(fp)
            return numpy.asarray(fingerprints), numpy.asarray(idx_fingerprints)

        def create_full_matrix():
            """Create matrix for all similarities."""
            similarity_matrix = numpy.zeros((len(references), len(queries)))
            if self.set_empty_scores == "nan":
                similarity_matrix[:] = numpy.nan
            elif isinstance(self.set_empty_scores, (float, int)):
                similarity_matrix[:] = self.set_empty_scores
            return similarity_matrix

        fingerprints1, idx_fingerprints1 = collect_fingerprints(references)
        fingerprints2, idx_fingerprints2 = collect_fingerprints(queries)
        assert idx_fingerprints1.size > 0 and idx_fingerprints2.size > 0, ("Not enouth molecular fingerprints.",
                                                                           "Apply 'add_fingerprint'filter first.")

        # Calculate similarity score matrix following specified method
        similarity_matrix = create_full_matrix()
        if self.similarity_measure == "jaccard":
            similarity_matrix[numpy.ix_(idx_fingerprints1,
                                        idx_fingerprints2)] = jaccard_similarity_matrix(fingerprints1,
                                                                                        fingerprints2)
        elif self.similarity_measure == "dice":
            similarity_matrix[numpy.ix_(idx_fingerprints1,
                                        idx_fingerprints2)] = dice_similarity_matrix(fingerprints1,
                                                                                     fingerprints2)
        elif self.similarity_measure == "cosine":
            similarity_matrix[numpy.ix_(idx_fingerprints1,
                                        idx_fingerprints2)] = cosine_similarity_matrix(fingerprints1,
                                                                                       fingerprints2)
        return similarity_matrix.astype(self.score_datatype)



import logging
from typing import Union
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def require_precursor_mz(spectrum_in: SpectrumType,
                         minimum_accepted_mz: float = 10.0
                         ) -> Union[SpectrumType, None]:

    """Returns None if there is no precursor_mz or if <= minimum_accepted_mz

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    minimum_accepted_mz:
        Set to minimum acceptable value for precursor m/z. Default is set to 10.0.
    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    precursor_mz = spectrum.get("precursor_mz", None)
    if precursor_mz is None:
        pepmass = spectrum.get("pepmass", None)
        assert pepmass is None or not isinstance(pepmass[0], (float, int)), \
            "Found 'pepmass' but no 'precursor_mz'. " \
            "Consider applying 'add_precursor_mz' filter first."
        return None

    assert isinstance(precursor_mz, (float, int)), \
        ("Expected 'precursor_mz' to be a scalar number.",
         "Consider applying 'add_precursor_mz' filter first.")
    if precursor_mz <= minimum_accepted_mz:
        logger.info("Spectrum without precursor_mz was set to None.")
        return None

    return spectrum



#!/usr/bin/env python
import os
from setuptools import find_packages
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(here, "matchms", "__version__.py")) as f:
    exec(f.read(), version)

with open("README.rst") as readme_file:
    readme = readme_file.read()

setup(
    name="matchms",
    version=version["__version__"],
    description="Python library for fuzzy comparison of mass spectrum data and other Python objects",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="Netherlands eScience Center",
    author_email="generalization@esciencecenter.nl",
    url="https://github.com/matchms/matchms",
    packages=find_packages(exclude=['*tests*']),
    package_data={"matchms": ["data/*.csv"]},
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords=[
        "similarity measures",
        "mass spectrometry",
        "fuzzy matching",
        "fuzzy search"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    test_suite="tests",
    python_requires='>=3.7',
    install_requires=[
        "deprecated",
        "lxml",
        "matplotlib",
        "networkx",
        "numba >=0.47",
        "numpy",
        "pyteomics >=4.2",
        "requests",
        "scipy",
    ],
    extras_require={"dev": ["bump2version",
                            "decorator",
                            "isort>=4.2.5,<5",
                            "pylint<2.12",
                            "prospector[with_pyroma]",
                            "pytest",
                            "pytest-cov",
                            "sphinx>=3.0.0,!=3.2.0,!=3.5.0,<4.0.0",
                            "sphinx_rtd_theme",
                            "sphinxcontrib-apidoc",
                            "testfixtures",
                            "yapf",],
                    "chemistry": ["rdkit >=2020.03.1"]},
)



from deprecated.sphinx import deprecated
from ..typing import SpectrumType
from .make_charge_int import make_charge_int


@deprecated(version='0.8.2', reason="Use expanded make_charge_int() instead.")
def make_charge_scalar(spectrum_in: SpectrumType) -> SpectrumType:
    """Convert charge field to scalar (if necessary).

    Deprecated function, now replaced by make_charge_int().
    """

    return make_charge_int(spectrum_in)



import logging
from math import ceil
from typing import Optional
from ..typing import SpectrumType


logger = logging.getLogger("matchms")


def require_minimum_number_of_peaks(spectrum_in: SpectrumType,
                                    n_required: int = 10,
                                    ratio_required: Optional[float] = None) -> SpectrumType:
    """Spectrum will be set to None when it has fewer peaks than required.

    Parameters
    ----------
    spectrum_in:
        Input spectrum.
    n_required:
        Number of minimum required peaks. Spectra with fewer peaks will be set
        to 'None'.
    ratio_required:
        Set desired ratio between minimum number of peaks and parent mass.
        Default is None.

    """
    if spectrum_in is None:
        return None

    spectrum = spectrum_in.clone()

    parent_mass = spectrum.get("parent_mass", None)
    if parent_mass and ratio_required:
        n_required_by_mass = int(ceil(ratio_required * parent_mass))
        threshold = max(n_required, n_required_by_mass)
    else:
        threshold = n_required

    if spectrum.peaks.intensities.size < threshold:
        logger.info("Spectrum with %s (<%s) peaks was set to None.",
                    str(spectrum.peaks.intensities.size), str(threshold))
        return None

    return spectrum
